# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
"""Wrapper to stack observations for single task environments."""

from collections import deque

import gym
import numpy as np

from mtenv.utils.types import ActionType, EnvStepReturnType


class FrameStack(gym.Wrapper):  # type: ignore[misc]
    # Mypy error: Class cannot subclass 'Wrapper' (has type 'Any')  [misc]

    def __init__(self, env: gym.core.Env, k: int):
        """Wrapper to stack observations for single task environments.

        Args:
            env (gym.core.Env): Single Task Environment
            k (int): number of frames to stack.
        """
        gym.Wrapper.__init__(self, env)
        self._k = k
        self._frames: deque = deque([], maxlen=k)
        shp = env.observation_space.shape
        self.observation_space = gym.spaces.Box(
            low=0,
            high=1,
            shape=((shp[0] * k,) + shp[1:]),
            dtype=env.observation_space.dtype,
        )
        self._max_episode_steps = env._max_episode_steps

    def reset(self) -> np.ndarray:
        obs = self.env.reset()
        for _ in range(self._k):
            self._frames.append(obs)
        return self._get_obs()

    def step(self, action: ActionType) -> EnvStepReturnType:
        obs, reward, done, info = self.env.step(action)
        self._frames.append(obs)
        return self._get_obs(), reward, done, info

    def _get_obs(self) -> np.ndarray:
        assert len(self._frames) == self._k
        return np.concatenate(list(self._frames), axis=0)



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Randomization functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from dm_control.mujoco.wrapper import mjbindings
import numpy as np
from six.moves import range


def random_limited_quaternion(random, limit):
    """Generates a random quaternion limited to the specified rotations."""
    axis = random.randn(3)
    axis /= np.linalg.norm(axis)
    angle = random.rand() * limit

    quaternion = np.zeros(4)
    mjbindings.mjlib.mju_axisAngle2Quat(quaternion, axis, angle)

    return quaternion


def randomize_limited_and_rotational_joints(physics, random=None):
    """Randomizes the positions of joints defined in the physics body.

    The following randomization rules apply:
      - Bounded joints (hinges or sliders) are sampled uniformly in the bounds.
      - Unbounded hinges are samples uniformly in [-pi, pi]
      - Quaternions for unlimited free joints and ball joints are sampled
        uniformly on the unit 3-sphere.
      - Quaternions for limited ball joints are sampled uniformly on a sector
        of the unit 3-sphere.
      - The linear degrees of freedom of free joints are not randomized.

    Args:
      physics: Instance of 'Physics' class that holds a loaded model.
      random: Optional instance of 'np.random.RandomState'. Defaults to the global
        NumPy random state.
    """
    random = random or np.random

    hinge = mjbindings.enums.mjtJoint.mjJNT_HINGE
    slide = mjbindings.enums.mjtJoint.mjJNT_SLIDE
    ball = mjbindings.enums.mjtJoint.mjJNT_BALL
    free = mjbindings.enums.mjtJoint.mjJNT_FREE

    qpos = physics.named.data.qpos

    for joint_id in range(physics.model.njnt):
        joint_name = physics.model.id2name(joint_id, "joint")
        joint_type = physics.model.jnt_type[joint_id]
        is_limited = physics.model.jnt_limited[joint_id]
        range_min, range_max = physics.model.jnt_range[joint_id]

        if is_limited:
            if joint_type == hinge or joint_type == slide:
                qpos[joint_name] = random.uniform(range_min, range_max)

            elif joint_type == ball:
                qpos[joint_name] = random_limited_quaternion(random, range_max)

        else:
            if joint_type == hinge:
                qpos[joint_name] = random.uniform(-np.pi, np.pi)

            elif joint_type == ball:
                quat = random.randn(4)
                quat /= np.linalg.norm(quat)
                qpos[joint_name] = quat

            elif joint_type == free:
                quat = random.rand(4)
                quat /= np.linalg.norm(quat)
                qpos[joint_name][3:] = quat



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
from typing import Any, Callable, Dict, List

from gym.core import Env

from mtenv import MTEnv
from mtenv.envs.hipbmdp import dmc_env
from mtenv.envs.shared.wrappers.multienv import MultiEnvWrapper

EnvBuilderType = Callable[[], Env]
TaskStateType = int
TaskObsType = int


def build(
    domain_name: str,
    task_name: str,
    seed: int,
    xml_file_ids: List[str],
    visualize_reward: bool,
    from_pixels: bool,
    height: int,
    width: int,
    frame_skip: int,
    frame_stack: int,
    sticky_observation_cfg: Dict[str, Any],
    initial_task_state: int = 1,
) -> MTEnv:
    """Build multitask environment as described in HiPBMDP paper. See
    :cite:`mtrl_as_a_hidden_block_mdp` for more details.

    Args:
        domain_name (str): name of the domain.
        task_name (str): name of the task.
        seed (int): environment seed (for reproducibility).
        xml_file_ids (List[str]): ids of xml files.
        visualize_reward (bool): should visualize reward ?
        from_pixels (bool): return pixel observations?
        height (int): height of pixel frames.
        width (int): width of pixel frames.
        frame_skip (int): should skip frames?
        frame_stack (int): should stack frames together?
        sticky_observation_cfg (Dict[str, Any]): Configuration for using
            sticky observations. It should be a dictionary with three
            keys, `should_use` which specifies if the config should be
            used, `sticky_probability` which specifies the probability of
            choosing a previous task and `last_k` which specifies the
            number of previous frames to choose from.
        initial_task_state (int, optional): intial task/environment
            to select. Defaults to 1.

    Returns:
        MTEnv:
    """

    def get_func_to_make_envs(xml_file_id: str) -> EnvBuilderType:
        def _func() -> Env:
            return dmc_env.build_dmc_env(
                domain_name=domain_name,
                task_name=task_name,
                seed=seed,
                xml_file_id=xml_file_id,
                visualize_reward=visualize_reward,
                from_pixels=from_pixels,
                height=height,
                width=width,
                frame_skip=frame_skip,
                frame_stack=frame_stack,
                sticky_observation_cfg=sticky_observation_cfg,
            )

        return _func

    funcs_to_make_envs = [
        get_func_to_make_envs(xml_file_id=file_id) for file_id in xml_file_ids
    ]

    mtenv = MultiEnvWrapper(
        funcs_to_make_envs=funcs_to_make_envs, initial_task_state=initial_task_state
    )
    return mtenv



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
from pathlib import Path

import setuptools

from mtenv.utils.setup_utils import parse_dependency

env_name = "control"
path = Path(__file__).parent / "requirements.txt"
requirements = parse_dependency(path)

with (Path(__file__).parent / "README.md").open() as fh:
    long_description = fh.read()

setuptools.setup(
    name=env_name,
    version="0.0.1",
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for dm_control.suite domains."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Internal dependencies.
from absl.testing import absltest
from absl.testing import parameterized
from dm_control import suite
from dm_control.rl import control
import mock
import numpy as np
import six
from six.moves import range
from six.moves import zip


def uniform_random_policy(action_spec, random=None):
    lower_bounds = action_spec.minimum
    upper_bounds = action_spec.maximum
    # Draw values between -1 and 1 for unbounded actions.
    lower_bounds = np.where(np.isinf(lower_bounds), -1.0, lower_bounds)
    upper_bounds = np.where(np.isinf(upper_bounds), 1.0, upper_bounds)
    random_state = np.random.RandomState(random)

    def policy(time_step):
        del time_step  # Unused.
        return random_state.uniform(lower_bounds, upper_bounds)

    return policy


def step_environment(env, policy, num_episodes=5, max_steps_per_episode=10):
    for _ in range(num_episodes):
        step_count = 0
        time_step = env.reset()
        yield time_step
        while not time_step.last():
            action = policy(time_step)
            time_step = env.step(action)
            step_count += 1
            yield time_step
            if step_count >= max_steps_per_episode:
                break


def make_trajectory(domain, task, seed, **trajectory_kwargs):
    env = suite.load(domain, task, task_kwargs={"random": seed})
    policy = uniform_random_policy(env.action_spec(), random=seed)
    return step_environment(env, policy, **trajectory_kwargs)


class DomainTest(parameterized.TestCase):
    """Tests run on all the tasks registered."""

    def test_constants(self):
        num_tasks = sum(len(tasks) for tasks in six.itervalues(suite.TASKS_BY_DOMAIN))

        self.assertLen(suite.ALL_TASKS, num_tasks)

    def _validate_observation(self, observation_dict, observation_spec):
        obs = observation_dict.copy()
        for name, spec in six.iteritems(observation_spec):
            arr = obs.pop(name)
            self.assertEqual(arr.shape, spec.shape)
            self.assertEqual(arr.dtype, spec.dtype)
            self.assertTrue(
                np.all(np.isfinite(arr)),
                msg="{!r} has non-finite value(s): {!r}".format(name, arr),
            )
        self.assertEmpty(
            obs,
            msg="Observation contains arrays(s) that are not in the spec: {!r}".format(
                obs
            ),
        )

    def _validate_reward_range(self, time_step):
        if time_step.first():
            self.assertIsNone(time_step.reward)
        else:
            self.assertIsInstance(time_step.reward, float)
            self.assertBetween(time_step.reward, 0, 1)

    def _validate_discount(self, time_step):
        if time_step.first():
            self.assertIsNone(time_step.discount)
        else:
            self.assertIsInstance(time_step.discount, float)
            self.assertBetween(time_step.discount, 0, 1)

    def _validate_control_range(self, lower_bounds, upper_bounds):
        for b in lower_bounds:
            self.assertEqual(b, -1.0)
        for b in upper_bounds:
            self.assertEqual(b, 1.0)

    @parameterized.parameters(*suite.ALL_TASKS)
    def test_components_have_names(self, domain, task):
        env = suite.load(domain, task)
        model = env.physics.model

        object_types_and_size_fields = [
            ("body", "nbody"),
            ("joint", "njnt"),
            ("geom", "ngeom"),
            ("site", "nsite"),
            ("camera", "ncam"),
            ("light", "nlight"),
            ("mesh", "nmesh"),
            ("hfield", "nhfield"),
            ("texture", "ntex"),
            ("material", "nmat"),
            ("equality", "neq"),
            ("tendon", "ntendon"),
            ("actuator", "nu"),
            ("sensor", "nsensor"),
            ("numeric", "nnumeric"),
            ("text", "ntext"),
            ("tuple", "ntuple"),
        ]
        for object_type, size_field in object_types_and_size_fields:
            for idx in range(getattr(model, size_field)):
                object_name = model.id2name(idx, object_type)
                self.assertNotEqual(
                    object_name,
                    "",
                    msg="Model {!r} contains unnamed {!r} with ID {}.".format(
                        model.name, object_type, idx
                    ),
                )

    @parameterized.parameters(*suite.ALL_TASKS)
    def test_model_has_at_least_2_cameras(self, domain, task):
        env = suite.load(domain, task)
        model = env.physics.model
        self.assertGreaterEqual(
            model.ncam,
            2,
            "Model {!r} should have at least 2 cameras, has {}.".format(
                model.name, model.ncam
            ),
        )

    @parameterized.parameters(*suite.ALL_TASKS)
    def test_task_conforms_to_spec(self, domain, task):
        """Tests that the environment timesteps conform to specifications."""
        is_benchmark = (domain, task) in suite.BENCHMARKING
        env = suite.load(domain, task)
        observation_spec = env.observation_spec()
        action_spec = env.action_spec()

        # Check action bounds.
        if is_benchmark:
            self._validate_control_range(action_spec.minimum, action_spec.maximum)

        # Step through the environment, applying random actions sampled within the
        # valid range and check the observations, rewards, and discounts.
        policy = uniform_random_policy(action_spec)
        for time_step in step_environment(env, policy):
            self._validate_observation(time_step.observation, observation_spec)
            self._validate_discount(time_step)
            if is_benchmark:
                self._validate_reward_range(time_step)

    @parameterized.parameters(*suite.ALL_TASKS)
    def test_environment_is_deterministic(self, domain, task):
        """Tests that identical seeds and actions produce identical trajectories."""
        seed = 0
        # Iterate over two trajectories generated using identical sequences of
        # random actions, and with identical task random states. Check that the
        # observations, rewards, discounts and step types are identical.
        trajectory1 = make_trajectory(domain=domain, task=task, seed=seed)
        trajectory2 = make_trajectory(domain=domain, task=task, seed=seed)
        for time_step1, time_step2 in zip(trajectory1, trajectory2):
            self.assertEqual(time_step1.step_type, time_step2.step_type)
            self.assertEqual(time_step1.reward, time_step2.reward)
            self.assertEqual(time_step1.discount, time_step2.discount)
            for key in six.iterkeys(time_step1.observation):
                np.testing.assert_array_equal(
                    time_step1.observation[key],
                    time_step2.observation[key],
                    err_msg="Observation {!r} is not equal.".format(key),
                )

    def assertCorrectColors(self, physics, reward):
        colors = physics.named.model.mat_rgba
        for material_name in ("self", "effector", "target"):
            highlight = colors[material_name + "_highlight"]
            default = colors[material_name + "_default"]
            blend_coef = reward ** 4
            expected = blend_coef * highlight + (1.0 - blend_coef) * default
            actual = colors[material_name]
            err_msg = (
                "Material {!r} has unexpected color.\nExpected: {!r}\n"
                "Actual: {!r}".format(material_name, expected, actual)
            )
            np.testing.assert_array_almost_equal(expected, actual, err_msg=err_msg)

    @parameterized.parameters(*suite.ALL_TASKS)
    def test_visualize_reward(self, domain, task):
        env = suite.load(domain, task)
        env.task.visualize_reward = True
        action = np.zeros(env.action_spec().shape)

        with mock.patch.object(env.task, "get_reward") as mock_get_reward:
            mock_get_reward.return_value = -3.0  # Rewards < 0 should be clipped.
            env.reset()
            mock_get_reward.assert_called_with(env.physics)
            self.assertCorrectColors(env.physics, reward=0.0)

            mock_get_reward.reset_mock()
            mock_get_reward.return_value = 0.5
            env.step(action)
            mock_get_reward.assert_called_with(env.physics)
            self.assertCorrectColors(env.physics, reward=mock_get_reward.return_value)

            mock_get_reward.reset_mock()
            mock_get_reward.return_value = 2.0  # Rewards > 1 should be clipped.
            env.step(action)
            mock_get_reward.assert_called_with(env.physics)
            self.assertCorrectColors(env.physics, reward=1.0)

            mock_get_reward.reset_mock()
            mock_get_reward.return_value = 0.25
            env.reset()
            mock_get_reward.assert_called_with(env.physics)
            self.assertCorrectColors(env.physics, reward=mock_get_reward.return_value)

    @parameterized.parameters(*suite.ALL_TASKS)
    def test_task_supports_environment_kwargs(self, domain, task):
        env = suite.load(domain, task, environment_kwargs=dict(flat_observation=True))
        # Check that the kwargs are actually passed through to the environment.
        self.assertSetEqual(set(env.observation_spec()), {control.FLAT_OBSERVATION_KEY})

    @parameterized.parameters(*suite.ALL_TASKS)
    def test_observation_arrays_dont_share_memory(self, domain, task):
        env = suite.load(domain, task)
        first_timestep = env.reset()
        action = np.zeros(env.action_spec().shape)
        second_timestep = env.step(action)
        for name, first_array in six.iteritems(first_timestep.observation):
            second_array = second_timestep.observation[name]
            self.assertFalse(
                np.may_share_memory(first_array, second_array),
                msg="Consecutive observations of {!r} may share memory.".format(name),
            )

    @parameterized.parameters(*suite.ALL_TASKS)
    def test_observations_dont_contain_constant_elements(self, domain, task):
        env = suite.load(domain, task)
        trajectory = make_trajectory(
            domain=domain, task=task, seed=0, num_episodes=2, max_steps_per_episode=1000
        )
        observations = {name: [] for name in env.observation_spec()}
        for time_step in trajectory:
            for name, array in six.iteritems(time_step.observation):
                observations[name].append(array)

        failures = []

        for name, array_list in six.iteritems(observations):
            # Sampling random uniform actions generally isn't sufficient to trigger
            # these touch sensors.
            if (
                domain in ("manipulator", "stacker")
                and name == "touch"
                or domain == "quadruped"
                and name == "force_torque"
            ):
                continue
            stacked_arrays = np.array(array_list)
            is_constant = np.all(stacked_arrays == stacked_arrays[0], axis=0)
            has_constant_elements = (
                is_constant if np.isscalar(is_constant) else np.any(is_constant)
            )
            if has_constant_elements:
                failures.append((name, is_constant))

        self.assertEmpty(
            failures,
            msg="The following observation(s) contain constant elements:\n{}".format(
                "\n".join(
                    ":\t".join([name, str(is_constant)])
                    for (name, is_constant) in failures
                )
            ),
        )

    @parameterized.parameters(*suite.ALL_TASKS)
    def test_initial_state_is_randomized(self, domain, task):
        env = suite.load(domain, task, task_kwargs={"random": 42})
        obs1 = env.reset().observation
        obs2 = env.reset().observation
        self.assertFalse(
            all(np.all(obs1[k] == obs2[k]) for k in obs1),
            "Two consecutive initial states have identical observations.\n"
            "First: {}\nSecond: {}".format(obs1, obs2),
        )


if __name__ == "__main__":
    absltest.main()



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Ball-in-Cup Domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from . import base
from . import common
from dm_control.utils import containers

_DEFAULT_TIME_LIMIT = 20  # (seconds)
_CONTROL_TIMESTEP = 0.02  # (seconds)


SUITE = containers.TaggedTasks()


def get_model_and_assets():
    """Returns a tuple containing the model XML string and a dict of assets."""
    return common.read_model("ball_in_cup.xml"), common.ASSETS


@SUITE.add("benchmarking", "easy")
def catch(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns the Ball-in-Cup task."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = BallInCup(random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        time_limit=time_limit,
        control_timestep=_CONTROL_TIMESTEP,
        **environment_kwargs
    )


class Physics(mujoco.Physics):
    """Physics with additional features for the Ball-in-Cup domain."""

    def ball_to_target(self):
        """Returns the vector from the ball to the target."""
        target = self.named.data.site_xpos["target", ["x", "z"]]
        ball = self.named.data.xpos["ball", ["x", "z"]]
        return target - ball

    def in_target(self):
        """Returns 1 if the ball is in the target, 0 otherwise."""
        ball_to_target = abs(self.ball_to_target())
        target_size = self.named.model.site_size["target", [0, 2]]
        ball_size = self.named.model.geom_size["ball", 0]
        return float(all(ball_to_target < target_size - ball_size))


class BallInCup(base.Task):
    """The Ball-in-Cup task. Put the ball in the cup."""

    def initialize_episode(self, physics):
        """Sets the state of the environment at the start of each episode.

        Args:
          physics: An instance of `Physics`.

        """
        # Find a collision-free random initial position of the ball.
        penetrating = True
        while penetrating:
            # Assign a random ball position.
            physics.named.data.qpos["ball_x"] = self.random.uniform(-0.2, 0.2)
            physics.named.data.qpos["ball_z"] = self.random.uniform(0.2, 0.5)
            # Check for collisions.
            physics.after_reset()
            penetrating = physics.data.ncon > 0
        super(BallInCup, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns an observation of the state."""
        obs = collections.OrderedDict()
        obs["position"] = physics.position()
        obs["velocity"] = physics.velocity()
        return obs

    def get_reward(self, physics):
        """Returns a sparse reward."""
        return physics.in_target()



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Fish Domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from . import base
from . import common
from dm_control.utils import containers
from dm_control.utils import rewards
import numpy as np


_DEFAULT_TIME_LIMIT = 40
_CONTROL_TIMESTEP = 0.04
_JOINTS = [
    "tail1",
    "tail_twist",
    "tail2",
    "finright_roll",
    "finright_pitch",
    "finleft_roll",
    "finleft_pitch",
]
SUITE = containers.TaggedTasks()


def get_model_and_assets():
    """Returns a tuple containing the model XML string and a dict of assets."""
    return common.read_model("fish.xml"), common.ASSETS


@SUITE.add("benchmarking")
def upright(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns the Fish Upright task."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Upright(random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        control_timestep=_CONTROL_TIMESTEP,
        time_limit=time_limit,
        **environment_kwargs
    )


@SUITE.add("benchmarking")
def swim(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns the Fish Swim task."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Swim(random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        control_timestep=_CONTROL_TIMESTEP,
        time_limit=time_limit,
        **environment_kwargs
    )


class Physics(mujoco.Physics):
    """Physics simulation with additional features for the Fish domain."""

    def upright(self):
        """Returns projection from z-axes of torso to the z-axes of worldbody."""
        return self.named.data.xmat["torso", "zz"]

    def torso_velocity(self):
        """Returns velocities and angular velocities of the torso."""
        return self.data.sensordata

    def joint_velocities(self):
        """Returns the joint velocities."""
        return self.named.data.qvel[_JOINTS]

    def joint_angles(self):
        """Returns the joint positions."""
        return self.named.data.qpos[_JOINTS]

    def mouth_to_target(self):
        """Returns a vector, from mouth to target in local coordinate of mouth."""
        data = self.named.data
        mouth_to_target_global = data.geom_xpos["target"] - data.geom_xpos["mouth"]
        return mouth_to_target_global.dot(data.geom_xmat["mouth"].reshape(3, 3))


class Upright(base.Task):
    """A Fish `Task` for getting the torso upright with smooth reward."""

    def __init__(self, random=None):
        """Initializes an instance of `Upright`.

        Args:
          random: Either an existing `numpy.random.RandomState` instance, an
            integer seed for creating a new `RandomState`, or None to select a seed
            automatically.
        """
        super(Upright, self).__init__(random=random)

    def initialize_episode(self, physics):
        """Randomizes the tail and fin angles and the orientation of the Fish."""
        quat = self.random.randn(4)
        physics.named.data.qpos["root"][3:7] = quat / np.linalg.norm(quat)
        for joint in _JOINTS:
            physics.named.data.qpos[joint] = self.random.uniform(-0.2, 0.2)
        # Hide the target. It's irrelevant for this task.
        physics.named.model.geom_rgba["target", 3] = 0
        super(Upright, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns an observation of joint angles, velocities and uprightness."""
        obs = collections.OrderedDict()
        obs["joint_angles"] = physics.joint_angles()
        obs["upright"] = physics.upright()
        obs["velocity"] = physics.velocity()
        return obs

    def get_reward(self, physics):
        """Returns a smooth reward."""
        return rewards.tolerance(physics.upright(), bounds=(1, 1), margin=1)


class Swim(base.Task):
    """A Fish `Task` for swimming with smooth reward."""

    def __init__(self, random=None):
        """Initializes an instance of `Swim`.

        Args:
          random: Optional, either a `numpy.random.RandomState` instance, an
            integer seed for creating a new `RandomState`, or None to select a seed
            automatically (default).
        """
        super(Swim, self).__init__(random=random)

    def initialize_episode(self, physics):
        """Sets the state of the environment at the start of each episode."""

        quat = self.random.randn(4)
        physics.named.data.qpos["root"][3:7] = quat / np.linalg.norm(quat)
        for joint in _JOINTS:
            physics.named.data.qpos[joint] = self.random.uniform(-0.2, 0.2)
        # Randomize target position.
        physics.named.model.geom_pos["target", "x"] = self.random.uniform(-0.4, 0.4)
        physics.named.model.geom_pos["target", "y"] = self.random.uniform(-0.4, 0.4)
        physics.named.model.geom_pos["target", "z"] = self.random.uniform(0.1, 0.3)
        super(Swim, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns an observation of joints, target direction and velocities."""
        obs = collections.OrderedDict()
        obs["joint_angles"] = physics.joint_angles()
        obs["upright"] = physics.upright()
        obs["target"] = physics.mouth_to_target()
        obs["velocity"] = physics.velocity()
        return obs

    def get_reward(self, physics):
        """Returns a smooth reward."""
        radii = physics.named.model.geom_size[["mouth", "target"], 0].sum()
        in_target = rewards.tolerance(
            np.linalg.norm(physics.mouth_to_target()),
            bounds=(0, radii),
            margin=2 * radii,
        )
        is_upright = 0.5 * (physics.upright() + 1)
        return (7 * in_target + is_upright) / 8



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
from typing import Any, Dict

import gym
from gym.core import Env
from gym.envs.registration import register

from mtenv.envs.hipbmdp.wrappers import framestack, sticky_observation


def _build_env(
    domain_name: str,
    task_name: str,
    seed: int = 1,
    xml_file_id: str = "none",
    visualize_reward: bool = True,
    from_pixels: bool = False,
    height: int = 84,
    width: int = 84,
    camera_id: int = 0,
    frame_skip: int = 1,
    environment_kwargs: Any = None,
    episode_length: int = 1000,
) -> Env:
    if xml_file_id is None:
        env_id = "dmc_%s_%s_%s-v1" % (domain_name, task_name, seed)
    else:
        env_id = "dmc_%s_%s_%s_%s-v1" % (domain_name, task_name, xml_file_id, seed)

    if from_pixels:
        assert (
            not visualize_reward
        ), "cannot use visualize reward when learning from pixels"

    # shorten episode length
    max_episode_steps = (episode_length + frame_skip - 1) // frame_skip

    if env_id not in gym.envs.registry.env_specs:
        register(
            id=env_id,
            entry_point="mtenv.envs.hipbmdp.wrappers.dmc_wrapper:DMCWrapper",
            kwargs={
                "domain_name": domain_name,
                "task_name": task_name,
                "task_kwargs": {"random": seed, "xml_file_id": xml_file_id},
                "environment_kwargs": environment_kwargs,
                "visualize_reward": visualize_reward,
                "from_pixels": from_pixels,
                "height": height,
                "width": width,
                "camera_id": camera_id,
                "frame_skip": frame_skip,
            },
            max_episode_steps=max_episode_steps,
        )
    return gym.make(env_id)


def build_dmc_env(
    domain_name: str,
    task_name: str,
    seed: int,
    xml_file_id: str,
    visualize_reward: bool,
    from_pixels: bool,
    height: int,
    width: int,
    frame_skip: int,
    frame_stack: int,
    sticky_observation_cfg: Dict[str, Any],
) -> Env:
    """Build a single DMC environment as described in
    :cite:`tassa2020dmcontrol`.

    Args:
        domain_name (str): name of the domain.
        task_name (str): name of the task.
        seed (int): environment seed (for reproducibility).
        xml_file_id (str): id of the xml file to use.
        visualize_reward (bool): should visualize reward ?
        from_pixels (bool): return pixel observations?
        height (int): height of pixel frames.
        width (int): width of pixel frames.
        frame_skip (int): should skip frames?
        frame_stack (int): should stack frames together?
        sticky_observation_cfg (Dict[str, Any]): Configuration for using
            sticky observations. It should be a dictionary with three
            keys, `should_use` which specifies if the config should be
            used, `sticky_probability` which specifies the probability of
            choosing a previous task and `last_k` which specifies the
            number of previous frames to choose from.

    Returns:
        Env:
    """
    env = _build_env(
        domain_name=domain_name,
        task_name=task_name,
        seed=seed,
        visualize_reward=visualize_reward,
        from_pixels=from_pixels,
        height=height,
        width=width,
        frame_skip=frame_skip,
        xml_file_id=xml_file_id,
    )
    if from_pixels:
        env = framestack.FrameStack(env, k=frame_stack)
    if sticky_observation_cfg and sticky_observation_cfg["should_use"]:
        env = sticky_observation.StickyObservation(  # type: ignore[attr-defined]
            env=env,
            sticky_probability=sticky_observation_cfg["sticky_probability"],
            last_k=sticky_observation_cfg["last_k"],
        )
    return env



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Demonstration of amc parsing for CMU mocap database.

To run the demo, supply a path to a `.amc` file:

    python mocap_demo --filename='path/to/mocap.amc'

CMU motion capture clips are available at mocap.cs.cmu.edu
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time

# Internal dependencies.

from absl import app
from absl import flags

from . import humanoid_CMU
from dm_control.suite.utils import parse_amc

import matplotlib.pyplot as plt
import numpy as np

FLAGS = flags.FLAGS
flags.DEFINE_string("filename", None, "amc file to be converted.")
flags.DEFINE_integer(
    "max_num_frames", 90, "Maximum number of frames for plotting/playback"
)


def main(unused_argv):
    env = humanoid_CMU.stand()

    # Parse and convert specified clip.
    converted = parse_amc.convert(FLAGS.filename, env.physics, env.control_timestep())

    max_frame = min(FLAGS.max_num_frames, converted.qpos.shape[1] - 1)

    width = 480
    height = 480
    video = np.zeros((max_frame, height, 2 * width, 3), dtype=np.uint8)

    for i in range(max_frame):
        p_i = converted.qpos[:, i]
        with env.physics.reset_context():
            env.physics.data.qpos[:] = p_i
        video[i] = np.hstack(
            [
                env.physics.render(height, width, camera_id=0),
                env.physics.render(height, width, camera_id=1),
            ]
        )

    tic = time.time()
    for i in range(max_frame):
        if i == 0:
            img = plt.imshow(video[i])
        else:
            img.set_data(video[i])
        toc = time.time()
        clock_dt = toc - tic
        tic = time.time()
        # Real-time playback not always possible as clock_dt > .03
        plt.pause(max(0.01, 0.03 - clock_dt))  # Need min display time > 0.0.
        plt.draw()
    plt.waitforbuttonpress()


if __name__ == "__main__":
    flags.mark_flag_as_required("filename")
    app.run(main)



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Hopper domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from . import base
from . import common
from dm_control.suite.utils import randomizers
from dm_control.utils import containers
from dm_control.utils import rewards
import numpy as np


SUITE = containers.TaggedTasks()

_CONTROL_TIMESTEP = 0.02  # (Seconds)

# Default duration of an episode, in seconds.
_DEFAULT_TIME_LIMIT = 20

# Minimal height of torso over foot above which stand reward is 1.
_STAND_HEIGHT = 0.6

# Hopping speed above which hop reward is 1.
_HOP_SPEED = 2


def get_model_and_assets():
    """Returns a tuple containing the model XML string and a dict of assets."""
    return common.read_model("hopper.xml"), common.ASSETS


@SUITE.add("benchmarking")
def stand(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns a Hopper that strives to stand upright, balancing its pose."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Hopper(hopping=False, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        time_limit=time_limit,
        control_timestep=_CONTROL_TIMESTEP,
        **environment_kwargs
    )


@SUITE.add("benchmarking")
def hop(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns a Hopper that strives to hop forward."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Hopper(hopping=True, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        time_limit=time_limit,
        control_timestep=_CONTROL_TIMESTEP,
        **environment_kwargs
    )


class Physics(mujoco.Physics):
    """Physics simulation with additional features for the Hopper domain."""

    def height(self):
        """Returns height of torso with respect to foot."""
        return self.named.data.xipos["torso", "z"] - self.named.data.xipos["foot", "z"]

    def speed(self):
        """Returns horizontal speed of the Hopper."""
        return self.named.data.sensordata["torso_subtreelinvel"][0]

    def touch(self):
        """Returns the signals from two foot touch sensors."""
        return np.log1p(self.named.data.sensordata[["touch_toe", "touch_heel"]])


class Hopper(base.Task):
    """A Hopper's `Task` to train a standing and a jumping Hopper."""

    def __init__(self, hopping, random=None):
        """Initialize an instance of `Hopper`.

        Args:
          hopping: Boolean, if True the task is to hop forwards, otherwise it is to
            balance upright.
          random: Optional, either a `numpy.random.RandomState` instance, an
            integer seed for creating a new `RandomState`, or None to select a seed
            automatically (default).
        """
        self._hopping = hopping
        super(Hopper, self).__init__(random=random)

    def initialize_episode(self, physics):
        """Sets the state of the environment at the start of each episode."""
        randomizers.randomize_limited_and_rotational_joints(physics, self.random)
        self._timeout_progress = 0
        super(Hopper, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns an observation of positions, velocities and touch sensors."""
        obs = collections.OrderedDict()
        # Ignores horizontal position to maintain translational invariance:
        obs["position"] = physics.data.qpos[1:].copy()
        obs["velocity"] = physics.velocity()
        obs["touch"] = physics.touch()
        return obs

    def get_reward(self, physics):
        """Returns a reward applicable to the performed task."""
        standing = rewards.tolerance(physics.height(), (_STAND_HEIGHT, 2))
        if self._hopping:
            hopping = rewards.tolerance(
                physics.speed(),
                bounds=(_HOP_SPEED, float("inf")),
                margin=_HOP_SPEED / 2,
                value_at_margin=0.5,
                sigmoid="linear",
            )
            return standing * hopping
        else:
            small_control = rewards.tolerance(
                physics.control(), margin=1, value_at_margin=0, sigmoid="quadratic"
            ).mean()
            small_control = (small_control + 4) / 5
            return standing * small_control



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
from pathlib import Path
from typing import List


def parse_dependency(filepath: Path) -> List[str]:
    """Parse python dependencies from a file.

    The list of dependencies is used by `setup.py` files. Lines starting
    with "#" are ingored (useful for writing comments). In case the
    dependnecy is host using git, the url is parsed and modified to make
    suitable for `setup.py` files.


    Args:
        filepath (Path):

    Returns:
        List[str]: List of dependencies
    """
    dep_list = []
    for dep in open(filepath).read().splitlines():
        if dep.startswith("#"):
            continue
        key = "#egg="
        if key in dep:
            git_link, egg_name = dep.split(key)
            dep = f"{egg_name} @ {git_link}"
        dep_list.append(dep)
    return dep_list



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""A collection of MuJoCo-based Reinforcement Learning environments."""

from __future__ import absolute_import, division, print_function

import collections
import inspect
import itertools
import sys

from dm_control.rl import control

from . import (
    acrobot,
    ball_in_cup,
    cartpole,
    cheetah,
    finger,
    fish,
    hopper,
    humanoid,
    humanoid_CMU,
    lqr,
    manipulator,
    pendulum,
    point_mass,
    quadruped,
    reacher,
    stacker,
    swimmer,
    walker,
)

# Find all domains imported.
_DOMAINS = {
    name: module
    for name, module in locals().items()
    if inspect.ismodule(module) and hasattr(module, "SUITE")
}


def _get_tasks(tag):
    """Returns a sequence of (domain name, task name) pairs for the given tag."""
    result = []

    for domain_name in sorted(_DOMAINS.keys()):
        domain = _DOMAINS[domain_name]

        if tag is None:
            tasks_in_domain = domain.SUITE
        else:
            tasks_in_domain = domain.SUITE.tagged(tag)

        for task_name in tasks_in_domain.keys():
            result.append((domain_name, task_name))

    return tuple(result)


def _get_tasks_by_domain(tasks):
    """Returns a dict mapping from task name to a tuple of domain names."""
    result = collections.defaultdict(list)

    for domain_name, task_name in tasks:
        result[domain_name].append(task_name)

    return {k: tuple(v) for k, v in result.items()}


# A sequence containing all (domain name, task name) pairs.
ALL_TASKS = _get_tasks(tag=None)

# Subsets of ALL_TASKS, generated via the tag mechanism.
BENCHMARKING = _get_tasks("benchmarking")
EASY = _get_tasks("easy")
HARD = _get_tasks("hard")
EXTRA = tuple(sorted(set(ALL_TASKS) - set(BENCHMARKING)))

# A mapping from each domain name to a sequence of its task names.
TASKS_BY_DOMAIN = _get_tasks_by_domain(ALL_TASKS)


def load(
    domain_name,
    task_name,
    task_kwargs=None,
    environment_kwargs=None,
    visualize_reward=False,
):
    """Returns an environment from a domain name, task name and optional settings.

    ```python
    env = suite.load('cartpole', 'balance')
    ```

    Args:
      domain_name: A string containing the name of a domain.
      task_name: A string containing the name of a task.
      task_kwargs: Optional `dict` of keyword arguments for the task.
      environment_kwargs: Optional `dict` specifying keyword arguments for the
        environment.
      visualize_reward: Optional `bool`. If `True`, object colours in rendered
        frames are set to indicate the reward at each step. Default `False`.

    Returns:
      The requested environment.
    """
    return build_environment(
        domain_name, task_name, task_kwargs, environment_kwargs, visualize_reward,
    )


def build_environment(
    domain_name,
    task_name,
    task_kwargs=None,
    environment_kwargs=None,
    visualize_reward=False,
):
    """Returns an environment from the suite given a domain name and a task name.

    Args:
      domain_name: A string containing the name of a domain.
      task_name: A string containing the name of a task.
      task_kwargs: Optional `dict` specifying keyword arguments for the task.
      environment_kwargs: Optional `dict` specifying keyword arguments for the
        environment.
      visualize_reward: Optional `bool`. If `True`, object colours in rendered
        frames are set to indicate the reward at each step. Default `False`.

    Raises:
      ValueError: If the domain or task doesn't exist.

    Returns:
      An instance of the requested environment.
    """
    if domain_name not in _DOMAINS:
        raise ValueError("Domain {!r} does not exist.".format(domain_name))

    domain = _DOMAINS[domain_name]

    if task_name not in domain.SUITE:
        raise ValueError(
            "Level {!r} does not exist in domain {!r}.".format(task_name, domain_name)
        )

    task_kwargs = task_kwargs or {}
    if environment_kwargs is not None:
        task_kwargs = task_kwargs.copy()
        task_kwargs["environment_kwargs"] = environment_kwargs
    env = domain.SUITE[task_name](**task_kwargs)
    env.task.visualize_reward = visualize_reward
    return env



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
"""Wrapper to enable sitcky observations for single task environments."""
# type: ignore
import random
from collections import deque

import gym


class StickyObservation(gym.Wrapper):
    def __init__(self, env: gym.Env, sticky_probability: float, last_k: int):
        """Env wrapper that returns a previous observation with probability
        `p` and the current observation with a probability `1-p`. `last_k`
        previous observations are stored.

        Args:
            env (gym.Env): Single task environment.
            sticky_probability (float): Probability `p` for returning a
                previous observation.
            last_k (int): Number of previous observations to store.

        Raises:
            ValueError: Raise a ValueError if `sticky_probability` is
                not in range `[0, 1]`.
        """
        super().__init__(self, env)
        if 1 >= sticky_probability >= 0:
            self._sticky_probability = sticky_probability
        else:
            raise ValueError(
                f"sticky_probability = {sticky_probability} is not in the interval [0, 1]."
            )
        self._last_k = last_k + 1
        self._observations: deque = deque([], maxlen=self._last_k)
        self.observation_space = env.observation_space
        self._max_episode_steps = env._max_episode_steps

    def reset(self):
        obs = self.env.reset()
        for _ in range(self._last_k):
            self._observations.append(obs)
        return self._get_obs()

    def step(self, action):
        obs, reward, done, info = self.env.step(action)
        self._observations.append(obs)
        return self._get_obs(), reward, done, info

    def _get_obs(self):
        assert len(self._observations) == self._last_k
        should_choose_old_observation = random.random() < self._sticky_probability
        if should_choose_old_observation:
            index = random.randint(0, self._last_k - 2)
            return self._observations[index]
        else:
            return self._observations[-1]



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import random
from typing import Any, Callable, Dict, List, Optional, Tuple

import metaworld
from gym import Env

from mtenv import MTEnv
from mtenv.envs.metaworld.wrappers.normalized_env import (  # type: ignore[attr-defined]
    NormalizedEnvWrapper,
)
from mtenv.envs.shared.wrappers.multienv import MultiEnvWrapper

EnvBuilderType = Callable[[], Env]
TaskStateType = int
TaskObsType = int
EnvIdToTaskMapType = Dict[str, metaworld.Task]


class MetaWorldMTWrapper(MultiEnvWrapper):
    def __init__(
        self,
        funcs_to_make_envs: List[EnvBuilderType],
        initial_task_state: TaskStateType,
        env_id_to_task_map: EnvIdToTaskMapType,
    ) -> None:
        """Wrapper to make MetaWorld environment compatible with Multitask
        Environment API.  See :cite:`yu2020meta` for more details about
        MetaWorld.

        Args:
            funcs_to_make_envs (List[EnvBuilderType]): list of constructor
                functions to make the environments.
            initial_task_state (TaskStateType): initial task/environment
                to select.
            env_id_to_task_map (EnvIdToTaskMapType): In MetaWorld, each
                environment can be associated with multiple tasks. This
                dict persists the mapping between environment ids and tasks.
        """
        super().__init__(
            funcs_to_make_envs=funcs_to_make_envs,
            initial_task_state=initial_task_state,
        )
        self.env_id_to_task_map = env_id_to_task_map


def get_list_of_func_to_make_envs(
    benchmark: Optional[metaworld.Benchmark],
    benchmark_name: str,
    env_id_to_task_map: Optional[EnvIdToTaskMapType],
    should_perform_reward_normalization: bool = True,
    task_name: str = "pick-place-v1",
    num_copies_per_env: int = 1,
) -> Tuple[List[Any], Dict[str, Any]]:
    """Return a list of functions to construct the MetaWorld environments
    and a mapping of environment ids to tasks.

    Args:
        benchmark (Optional[metaworld.Benchmark]): `benchmark` to create
            tasks from.
        benchmark_name (str): name of the `benchmark`. This is used only
            when the `benchmark` is None.
        env_id_to_task_map (Optional[EnvIdToTaskMapType]): In MetaWorld,
            each environment can be associated with multiple tasks. This
            dict persists the mapping between environment ids and tasks.
        should_perform_reward_normalization (bool, optional): Defaults to
            True.
        task_name (str, optional): In case of MT1, only . Defaults to
            "pick-place-v1".
        num_copies_per_env (int, optional): Number of copies to create for
            each environment. Defaults to 1.

    Raises:
        ValueError: if `benchmark` is None and `benchmark_name` is not
            MT1, MT10, or MT50.

    Returns:
        Tuple[List[Any], Dict[str, Any]]: A tuple of two elements. The
        first element is a list of functions to construct the MetaWorld
        environments and the second is a mapping of environment ids
        to tasks.

    """
    if not benchmark:
        if benchmark_name == "MT1":
            benchmark = metaworld.ML1(task_name)
        elif benchmark_name == "MT10":
            benchmark = metaworld.MT10()
        elif benchmark_name == "MT50":
            benchmark = metaworld.MT50()
        else:
            raise ValueError(f"benchmark_name={benchmark_name} is not valid.")

    env_id_list = list(benchmark.train_classes.keys())

    def _get_class_items(current_benchmark):
        return current_benchmark.train_classes.items()

    def _get_tasks(current_benchmark):
        return current_benchmark.train_tasks

    def _get_env_id_to_task_map() -> EnvIdToTaskMapType:
        env_id_to_task_map: EnvIdToTaskMapType = {}
        current_benchmark = benchmark
        for env_id in env_id_list:
            for name, _ in _get_class_items(current_benchmark):
                if name == env_id:
                    task = random.choice(
                        [
                            task
                            for task in _get_tasks(current_benchmark)
                            if task.env_name == name
                        ]
                    )
                    env_id_to_task_map[env_id] = task
        return env_id_to_task_map

    if env_id_to_task_map is None:
        env_id_to_task_map: EnvIdToTaskMapType = _get_env_id_to_task_map()  # type: ignore[no-redef]
    assert env_id_to_task_map is not None

    def get_func_to_make_envs(env_id: str):
        current_benchmark = benchmark

        def _make_env():
            for name, env_cls in _get_class_items(current_benchmark):
                if name == env_id:
                    env = env_cls()
                    task = env_id_to_task_map[env_id]
                    env.set_task(task)
                    if should_perform_reward_normalization:
                        env = NormalizedEnvWrapper(env, normalize_reward=True)
                    return env

        return _make_env

    if num_copies_per_env > 1:
        env_id_list = [
            [env_id for _ in range(num_copies_per_env)] for env_id in env_id_list
        ]
        env_id_list = [
            env_id for env_id_sublist in env_id_list for env_id in env_id_sublist
        ]

    funcs_to_make_envs = [get_func_to_make_envs(env_id) for env_id in env_id_list]

    return funcs_to_make_envs, env_id_to_task_map


def build(
    benchmark: Optional[metaworld.Benchmark],
    benchmark_name: str,
    env_id_to_task_map: Optional[EnvIdToTaskMapType],
    should_perform_reward_normalization: bool = True,
    task_name: str = "pick-place-v1",
    num_copies_per_env: int = 1,
    initial_task_state: int = 1,
) -> MTEnv:
    """Build a MTEnv comptaible variant of MetaWorld.

    Args:
        benchmark (Optional[metaworld.Benchmark]): `benchmark` to create
            tasks from.
        benchmark_name (str): name of the `benchmark`. This is used only
            when the `benchmark` is None.
        env_id_to_task_map (Optional[EnvIdToTaskMapType]): In MetaWorld,
            each environment can be associated with multiple tasks. This
            dict persists the mapping between environment ids and tasks.
        should_perform_reward_normalization (bool, optional): Defaults to
            True.
        task_name (str, optional): In case of MT1, only . Defaults to
            "pick-place-v1".
        num_copies_per_env (int, optional): Number of copies to create for
            each environment. Defaults to 1.
        initial_task_state (int, optional): initial task/environment to
            select. Defaults to 1.

    Returns:
        MTEnv:
    """
    funcs_to_make_envs, env_id_to_task_map = get_list_of_func_to_make_envs(
        benchmark=benchmark,
        benchmark_name=benchmark_name,
        env_id_to_task_map=env_id_to_task_map,
        should_perform_reward_normalization=should_perform_reward_normalization,
        task_name=task_name,
        num_copies_per_env=num_copies_per_env,
    )

    assert env_id_to_task_map is not None

    mtenv = MetaWorldMTWrapper(
        funcs_to_make_envs=funcs_to_make_envs,
        initial_task_state=initial_task_state,
        env_id_to_task_map=env_id_to_task_map,
    )
    return mtenv



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
"""Wrapper to change the behaviour of an existing multitask environment."""

from typing import List, Optional

from numpy.random import RandomState

from mtenv import MTEnv
from mtenv.utils import seeding
from mtenv.utils.types import (
    ActionType,
    ObsType,
    StepReturnType,
    TaskObsType,
    TaskStateType,
)


class MultiTask(MTEnv):
    def __init__(self, env: MTEnv):
        """Wrapper to change the behaviour of an existing multitask environment

        Args:
            env (MTEnv): Multitask environment to wrap over.
        """
        self.env = env
        self.observation_space = self.env.observation_space
        self.action_space = self.env.action_space
        self.np_random_env: Optional[RandomState] = None
        self.np_random_task: Optional[RandomState] = None

    def step(self, action: ActionType) -> StepReturnType:
        return self.env.step(action)

    def get_task_obs(self) -> TaskObsType:
        return self.env.get_task_obs()

    def get_task_state(self) -> TaskStateType:
        return self.env.get_task_state()

    def set_task_state(self, task_state: TaskStateType) -> None:
        self.env.set_task_state(task_state)

    def assert_env_seed_is_set(self) -> None:
        """Check that the env seed is set."""
        assert self.np_random_env is not None, "please call `seed()` first"
        self.env.assert_env_seed_is_set()

    def assert_task_seed_is_set(self) -> None:
        """Check that the task seed is set."""
        assert self.np_random_task is not None, "please call `seed_task()` first"
        self.env.assert_task_seed_is_set()

    def reset(self) -> ObsType:
        return self.env.reset()

    def sample_task_state(self) -> TaskStateType:
        return self.env.sample_task_state()

    def reset_task_state(self) -> None:
        self.env.reset_task_state()

    def seed(self, seed: Optional[int] = None) -> List[int]:
        self.np_random_env, seed = seeding.np_random(seed)
        return [seed] + self.env.seed(seed)

    def seed_task(self, seed: Optional[int] = None) -> List[int]:
        self.np_random_task, seed = seeding.np_random(seed)
        return [seed] + self.env.seed_task(seed)



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Planar Manipulator domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from . import base
from . import common
from dm_control.utils import containers
from dm_control.utils import rewards
from dm_control.utils import xml_tools

from lxml import etree
import numpy as np

_CLOSE = 0.01  # (Meters) Distance below which a thing is considered close.
_CONTROL_TIMESTEP = 0.01  # (Seconds)
_TIME_LIMIT = 10  # (Seconds)
_P_IN_HAND = 0.1  # Probabillity of object-in-hand initial state
_P_IN_TARGET = 0.1  # Probabillity of object-in-target initial state
_ARM_JOINTS = [
    "arm_root",
    "arm_shoulder",
    "arm_elbow",
    "arm_wrist",
    "finger",
    "fingertip",
    "thumb",
    "thumbtip",
]
_ALL_PROPS = frozenset(["ball", "target_ball", "cup", "peg", "target_peg", "slot"])

SUITE = containers.TaggedTasks()


def make_model(use_peg, insert):
    """Returns a tuple containing the model XML string and a dict of assets."""
    xml_string = common.read_model("manipulator.xml")
    parser = etree.XMLParser(remove_blank_text=True)
    mjcf = etree.XML(xml_string, parser)

    # Select the desired prop.
    if use_peg:
        required_props = ["peg", "target_peg"]
        if insert:
            required_props += ["slot"]
    else:
        required_props = ["ball", "target_ball"]
        if insert:
            required_props += ["cup"]

    # Remove unused props
    for unused_prop in _ALL_PROPS.difference(required_props):
        prop = xml_tools.find_element(mjcf, "body", unused_prop)
        prop.getparent().remove(prop)

    return etree.tostring(mjcf, pretty_print=True), common.ASSETS


@SUITE.add("benchmarking", "hard")
def bring_ball(
    fully_observable=True, time_limit=_TIME_LIMIT, random=None, environment_kwargs=None
):
    """Returns manipulator bring task with the ball prop."""
    use_peg = False
    insert = False
    physics = Physics.from_xml_string(*make_model(use_peg, insert))
    task = Bring(
        use_peg=use_peg, insert=insert, fully_observable=fully_observable, random=random
    )
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        control_timestep=_CONTROL_TIMESTEP,
        time_limit=time_limit,
        **environment_kwargs
    )


@SUITE.add("hard")
def bring_peg(
    fully_observable=True, time_limit=_TIME_LIMIT, random=None, environment_kwargs=None
):
    """Returns manipulator bring task with the peg prop."""
    use_peg = True
    insert = False
    physics = Physics.from_xml_string(*make_model(use_peg, insert))
    task = Bring(
        use_peg=use_peg, insert=insert, fully_observable=fully_observable, random=random
    )
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        control_timestep=_CONTROL_TIMESTEP,
        time_limit=time_limit,
        **environment_kwargs
    )


@SUITE.add("hard")
def insert_ball(
    fully_observable=True, time_limit=_TIME_LIMIT, random=None, environment_kwargs=None
):
    """Returns manipulator insert task with the ball prop."""
    use_peg = False
    insert = True
    physics = Physics.from_xml_string(*make_model(use_peg, insert))
    task = Bring(
        use_peg=use_peg, insert=insert, fully_observable=fully_observable, random=random
    )
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        control_timestep=_CONTROL_TIMESTEP,
        time_limit=time_limit,
        **environment_kwargs
    )


@SUITE.add("hard")
def insert_peg(
    fully_observable=True, time_limit=_TIME_LIMIT, random=None, environment_kwargs=None
):
    """Returns manipulator insert task with the peg prop."""
    use_peg = True
    insert = True
    physics = Physics.from_xml_string(*make_model(use_peg, insert))
    task = Bring(
        use_peg=use_peg, insert=insert, fully_observable=fully_observable, random=random
    )
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        control_timestep=_CONTROL_TIMESTEP,
        time_limit=time_limit,
        **environment_kwargs
    )


class Physics(mujoco.Physics):
    """Physics with additional features for the Planar Manipulator domain."""

    def bounded_joint_pos(self, joint_names):
        """Returns joint positions as (sin, cos) values."""
        joint_pos = self.named.data.qpos[joint_names]
        return np.vstack([np.sin(joint_pos), np.cos(joint_pos)]).T

    def joint_vel(self, joint_names):
        """Returns joint velocities."""
        return self.named.data.qvel[joint_names]

    def body_2d_pose(self, body_names, orientation=True):
        """Returns positions and/or orientations of bodies."""
        if not isinstance(body_names, str):
            body_names = np.array(body_names).reshape(-1, 1)  # Broadcast indices.
        pos = self.named.data.xpos[body_names, ["x", "z"]]
        if orientation:
            ori = self.named.data.xquat[body_names, ["qw", "qy"]]
            return np.hstack([pos, ori])
        else:
            return pos

    def touch(self):
        return np.log1p(self.data.sensordata)

    def site_distance(self, site1, site2):
        site1_to_site2 = np.diff(self.named.data.site_xpos[[site2, site1]], axis=0)
        return np.linalg.norm(site1_to_site2)


class Bring(base.Task):
    """A Bring `Task`: bring the prop to the target."""

    def __init__(self, use_peg, insert, fully_observable, random=None):
        """Initialize an instance of the `Bring` task.

        Args:
          use_peg: A `bool`, whether to replace the ball prop with the peg prop.
          insert: A `bool`, whether to insert the prop in a receptacle.
          fully_observable: A `bool`, whether the observation should contain the
            position and velocity of the object being manipulated and the target
            location.
          random: Optional, either a `numpy.random.RandomState` instance, an
            integer seed for creating a new `RandomState`, or None to select a seed
            automatically (default).
        """
        self._use_peg = use_peg
        self._target = "target_peg" if use_peg else "target_ball"
        self._object = "peg" if self._use_peg else "ball"
        self._object_joints = ["_".join([self._object, dim]) for dim in "xzy"]
        self._receptacle = "slot" if self._use_peg else "cup"
        self._insert = insert
        self._fully_observable = fully_observable
        super(Bring, self).__init__(random=random)

    def initialize_episode(self, physics):
        """Sets the state of the environment at the start of each episode."""
        # Local aliases
        choice = self.random.choice
        uniform = self.random.uniform
        model = physics.named.model
        data = physics.named.data

        # Find a collision-free random initial configuration.
        penetrating = True
        while penetrating:

            # Randomise angles of arm joints.
            is_limited = model.jnt_limited[_ARM_JOINTS].astype(np.bool)
            joint_range = model.jnt_range[_ARM_JOINTS]
            lower_limits = np.where(is_limited, joint_range[:, 0], -np.pi)
            upper_limits = np.where(is_limited, joint_range[:, 1], np.pi)
            angles = uniform(lower_limits, upper_limits)
            data.qpos[_ARM_JOINTS] = angles

            # Symmetrize hand.
            data.qpos["finger"] = data.qpos["thumb"]

            # Randomise target location.
            target_x = uniform(-0.4, 0.4)
            target_z = uniform(0.1, 0.4)
            if self._insert:
                target_angle = uniform(-np.pi / 3, np.pi / 3)
                model.body_pos[self._receptacle, ["x", "z"]] = target_x, target_z
                model.body_quat[self._receptacle, ["qw", "qy"]] = [
                    np.cos(target_angle / 2),
                    np.sin(target_angle / 2),
                ]
            else:
                target_angle = uniform(-np.pi, np.pi)

            model.body_pos[self._target, ["x", "z"]] = target_x, target_z
            model.body_quat[self._target, ["qw", "qy"]] = [
                np.cos(target_angle / 2),
                np.sin(target_angle / 2),
            ]

            # Randomise object location.
            object_init_probs = [
                _P_IN_HAND,
                _P_IN_TARGET,
                1 - _P_IN_HAND - _P_IN_TARGET,
            ]
            init_type = choice(["in_hand", "in_target", "uniform"], p=object_init_probs)
            if init_type == "in_target":
                object_x = target_x
                object_z = target_z
                object_angle = target_angle
            elif init_type == "in_hand":
                physics.after_reset()
                object_x = data.site_xpos["grasp", "x"]
                object_z = data.site_xpos["grasp", "z"]
                grasp_direction = data.site_xmat["grasp", ["xx", "zx"]]
                object_angle = np.pi - np.arctan2(
                    grasp_direction[1], grasp_direction[0]
                )
            else:
                object_x = uniform(-0.5, 0.5)
                object_z = uniform(0, 0.7)
                object_angle = uniform(0, 2 * np.pi)
                data.qvel[self._object + "_x"] = uniform(-5, 5)

            data.qpos[self._object_joints] = object_x, object_z, object_angle

            # Check for collisions.
            physics.after_reset()
            penetrating = physics.data.ncon > 0

        super(Bring, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns either features or only sensors (to be used with pixels)."""
        obs = collections.OrderedDict()
        obs["arm_pos"] = physics.bounded_joint_pos(_ARM_JOINTS)
        obs["arm_vel"] = physics.joint_vel(_ARM_JOINTS)
        obs["touch"] = physics.touch()
        if self._fully_observable:
            obs["hand_pos"] = physics.body_2d_pose("hand")
            obs["object_pos"] = physics.body_2d_pose(self._object)
            obs["object_vel"] = physics.joint_vel(self._object_joints)
            obs["target_pos"] = physics.body_2d_pose(self._target)
        return obs

    def _is_close(self, distance):
        return rewards.tolerance(distance, (0, _CLOSE), _CLOSE * 2)

    def _peg_reward(self, physics):
        """Returns a reward for bringing the peg prop to the target."""
        grasp = self._is_close(physics.site_distance("peg_grasp", "grasp"))
        pinch = self._is_close(physics.site_distance("peg_pinch", "pinch"))
        grasping = (grasp + pinch) / 2
        bring = self._is_close(physics.site_distance("peg", "target_peg"))
        bring_tip = self._is_close(physics.site_distance("target_peg_tip", "peg_tip"))
        bringing = (bring + bring_tip) / 2
        return max(bringing, grasping / 3)

    def _ball_reward(self, physics):
        """Returns a reward for bringing the ball prop to the target."""
        return self._is_close(physics.site_distance("ball", "target_ball"))

    def get_reward(self, physics):
        """Returns a reward to the agent."""
        if self._use_peg:
            return self._peg_reward(physics)
        else:
            return self._ball_reward(physics)



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for the dm_control.suite loader."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Internal dependencies.

from absl.testing import absltest

from dm_control import suite
from dm_control.rl import control


class LoaderTest(absltest.TestCase):
    def test_load_without_kwargs(self):
        env = suite.load("cartpole", "swingup")
        self.assertIsInstance(env, control.Environment)

    def test_load_with_kwargs(self):
        env = suite.load(
            "cartpole", "swingup", task_kwargs={"time_limit": 40, "random": 99}
        )
        self.assertIsInstance(env, control.Environment)


class LoaderConstantsTest(absltest.TestCase):
    def testSuiteConstants(self):
        self.assertNotEmpty(suite.BENCHMARKING)
        self.assertNotEmpty(suite.EASY)
        self.assertNotEmpty(suite.HARD)
        self.assertNotEmpty(suite.EXTRA)


if __name__ == "__main__":
    absltest.main()



# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the  LICENSE file in the root directory of this source tree.

import copy
import math
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from gym import spaces
from gym.spaces.box import Box as BoxSpace
from gym.spaces.dict import Dict as DictSpace
from gym.spaces.discrete import Discrete as DiscreteSpace
from gym_miniworld.entity import Box
from gym_miniworld.miniworld import Agent, MiniWorldEnv
from numpy.random import RandomState

from mtenv.utils import seeding
from mtenv.utils.types import DoneType, InfoType, RewardType, TaskObsType
from mtenv.wrappers.env_to_mtenv import EnvToMTEnv

TaskStateType = List[int]

ActionType = int

EnvObsType = Dict[str, Union[int, List[int], List[float]]]
ObsType = Dict[str, Union[EnvObsType, TaskObsType]]
StepReturnType = Tuple[ObsType, RewardType, DoneType, InfoType]


class MTMiniWorldEnv(EnvToMTEnv):
    def make_observation(self, env_obs: EnvObsType) -> ObsType:
        raise NotImplementedError

    def get_task_obs(self) -> TaskObsType:
        return self.env.get_task_obs()

    def get_task_state(self) -> TaskStateType:
        return self.env.task_state

    def set_task_state(self, task_state: TaskStateType) -> None:
        self.env.set_task_state(task_state)

    def sample_task_state(self) -> TaskStateType:
        return self.env.sample_task_state()

    def reset(self, **kwargs: Dict[str, Any]) -> ObsType:  # type: ignore[override]
        # signature is incompatible with supertype.
        self.assert_env_seed_is_set()
        return self.env.reset(**kwargs)

    def step(self, action: ActionType) -> StepReturnType:  # type: ignore
        return self.env.step(action)

    def assert_env_seed_is_set(self) -> None:
        assert self.env.np_random_env is not None, "please call `seed()` first"

    def assert_task_seed_is_set(self) -> None:
        assert self.env.np_random_task is not None, "please call `seed_task()` first"

    def seed(self, seed: Optional[int] = None) -> List[int]:
        """Set the seed for environment observations"""
        return self.env.seed(seed=seed)


class TwoGoalMazeEnv(MiniWorldEnv):

    metadata = {"render.modes": ["human", "rgb_array"], "video.frames_per_second": 30}

    def __init__(
        self,
        size_x=5,
        size_y=5,
        obs_type="xy",
        task_seed=0,
        n_tasks=10,
        p_change=0.0,
        empty_mu=False,
    ):
        assert p_change == 0.0
        self.empty_mu = empty_mu
        self.obs_type = obs_type
        self.seed_task(seed=task_seed)
        self.np_random_env: Optional[RandomState] = None
        self.size_x, self.size_y = size_x, size_y
        self.task_state = []

        super().__init__()
        # Allow only movement actions (left/right/forward)
        self.action_space = spaces.Discrete(self.actions.move_forward + 1)
        if self.obs_type == "xy":
            _obs_space = BoxSpace(
                low=-np.inf, high=np.inf, shape=(4,), dtype=np.float32
            )

        else:
            _obs_space = BoxSpace(
                low=-1.0,
                high=1.0,
                shape=(64, 64),
                dtype=np.float32,
            )
        self.observation_space = DictSpace(
            {
                "obs": _obs_space,
                "total_reward": BoxSpace(
                    low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
                ),
            }
        )

    def assert_env_seed_is_set(self) -> None:
        """Check that the env seed is set."""
        assert self.np_random_env is not None, "please call `seed()` first"

    def assert_task_seed_is_set(self) -> None:
        """Check that the task seed is set."""
        assert self.np_random_task is not None, "please call `seed_task()` first"

    def seed_task(self, seed: Optional[int] = None) -> List[int]:
        """Set the seed for task information"""
        self.np_random_task, seed = seeding.np_random(seed)
        assert isinstance(seed, int)
        return [seed]

    def sample_task_state(self) -> TaskStateType:
        self.assert_task_seed_is_set()
        return [self.np_random_task.randint(2)]

    def set_task_state(self, task_state: TaskStateType) -> None:
        self.task_state = task_state

    def _gen_world(self):
        self.reset_task_state()
        room1 = self.add_rect_room(
            min_x=-self.size_x,
            max_x=self.size_x,
            min_z=-self.size_y,
            max_z=self.size_y,
            wall_tex="brick_wall",
        )
        self.room1 = room1
        room2 = self.add_rect_room(
            min_x=-self.size_x,
            max_x=self.size_x,
            min_z=self.size_y,
            max_z=self.size_y + 1,
            wall_tex="cardboard",
        )
        self.connect_rooms(room1, room2, min_x=-self.size_x, max_x=self.size_x)

        room3 = self.add_rect_room(
            min_x=-self.size_x,
            max_x=self.size_x,
            min_z=-self.size_y - 1,
            max_z=-self.size_y,
            wall_tex="lava",
        )
        self.connect_rooms(room1, room3, min_x=-self.size_x, max_x=self.size_x)

        room4 = None
        if self.task_state[0] == 0:
            room4 = self.add_rect_room(
                min_x=-self.size_x - 1,
                max_x=-self.size_x,
                min_z=-self.size_y,
                max_z=self.size_y,
                wall_tex="wood_planks",
            )
        else:
            room4 = self.add_rect_room(
                min_x=-self.size_x - 1,
                max_x=-self.size_x,
                min_z=-self.size_y,
                max_z=self.size_y,
                wall_tex="slime",
            )

        self.connect_rooms(room1, room4, min_z=-self.size_y, max_z=self.size_y)

        room5 = self.add_rect_room(
            min_x=self.size_x,
            max_x=self.size_x + 1,
            min_z=-self.size_y,
            max_z=self.size_y,
            wall_tex="metal_grill",
        )

        self.connect_rooms(room1, room5, min_z=-self.size_y, max_z=self.size_y)

        self.boxes = []
        self.boxes.append(Box(color="blue"))
        self.boxes.append(Box(color="red"))
        self.place_entity(self.boxes[0], room=room1)
        self.place_entity(self.boxes[1], room=room1)

        # Choose a random room and position to spawn at
        _dir = self.np_random_env.randint(8) * (math.pi / 4) - math.pi
        self.place_agent(
            dir=_dir,
            room=room1,
        )
        while self._dist() < 2 or self._ndist() < 2:
            self.place_agent(
                dir=_dir,
                room=room1,
            )

    def _dist(self):
        bp = self.boxes[int(self.task_state[0])].pos
        pos = self.agent.pos
        distance = math.sqrt((bp[0] - pos[0]) ** 2 + (bp[2] - pos[2]) ** 2)

        return distance

    def _ndist(self):
        bp = self.boxes[1 - int(self.task_state[0])].pos
        pos = self.agent.pos
        distance = math.sqrt((bp[0] - pos[0]) ** 2 + (bp[2] - pos[2]) ** 2)

        return distance

    def reset(self) -> ObsType:
        self.assert_env_seed_is_set()
        self.max_episode_steps = 200
        self.treward = 0.0
        self.step_count = 0
        self.agent = Agent()
        self.entities: List[Any] = []
        self.rooms: List[Any] = []
        self.wall_segs: List[Any] = []
        self._gen_world()
        self.blocked = False
        rand = self.rand if self.domain_rand else None
        self.params.sample_many(
            rand, self, ["sky_color", "light_pos", "light_color", "light_ambient"]
        )

        for ent in self.entities:
            ent.randomize(self.params, rand)

        # Compute the min and max x, z extents of the whole floorplan
        self.min_x = min(r.min_x for r in self.rooms)
        self.max_x = max(r.max_x for r in self.rooms)
        self.min_z = min(r.min_z for r in self.rooms)
        self.max_z = max(r.max_z for r in self.rooms)

        # Generate static data
        if len(self.wall_segs) == 0:
            self._gen_static_data()

        # Pre-compile static parts of the environment into a display list
        self._render_static()
        _pos = [
            (self.agent.pos[0] / self.size_x) * 2.1 - 1.0,
            (self.agent.pos[2] / self.size_y) * 2.1 - 1.0,
        ]
        _dir = [self.agent.dir_vec[0], self.agent.dir_vec[2]]

        if self.obs_type == "xy":
            _mu = [0.0]
            at = math.atan2(_dir[0], _dir[1])
            o = copy.deepcopy(_pos + [at] + _mu)
        else:
            o = (self.render_obs() / 255.0) * 2.0 - 1.0

        return self.make_obs(env_obs=o, total_reward=[0.0])

    def get_task_obs(self) -> TaskObsType:
        mmu = copy.deepcopy(self.task_state)
        if self.empty_mu:
            mmu = [0.0]
        return mmu

    def get_task_state(self) -> TaskStateType:
        return self.task_state

    def reset_task_state(self) -> None:
        """Sample a new task_state and set that as the new task_state"""
        self.set_task_state(task_state=self.sample_task_state())

    def make_obs(self, env_obs: Any, total_reward: List[float]) -> ObsType:

        return {
            "env_obs": {"obs": env_obs, "total_reward": total_reward},
            "task_obs": self.get_task_obs(),
        }

    def seed(self, seed: Optional[int] = None) -> List[int]:
        """Set the seed for environment observations"""
        self.np_random_env, seed = seeding.np_random(seed)
        return [seed] + super().seed(seed=seed)

    def step(self, action: ActionType) -> StepReturnType:
        self.step_count += 1
        if not self.blocked:
            if action == 2:
                self.move_agent(0.51, 0.0)  # fwd_step, fwd_drift)
            elif action == 0:
                self.turn_agent(45)
            elif action == 1:
                self.turn_agent(-45)
        reward = 0.0
        done = False

        distance = self._dist()
        if distance < 2:
            reward = +1.0
            done = True
        distance = self._ndist()
        if distance < 2:
            reward = -1.0
            done = True
        _pos = [
            (self.agent.pos[0] / self.size_x) * 2.1 - 1.0,
            (self.agent.pos[2] / self.size_y) * 2.1 - 1.0,
        ]
        _dir = [self.agent.dir_vec[0], self.agent.dir_vec[2]]

        if self.obs_type == "xy":
            at = math.atan2(_dir[0], _dir[1])
            _mu = [0.0]
            if (at < -1.5 and at > -1.7) and not self.empty_mu:
                _mu = [1.0]
                if self.task_state[0] == 0:
                    _mu = [-1.0]

            o = copy.deepcopy(_pos + [at] + _mu)
        else:
            o = (self.render_obs() / 255.0) * 2.0 - 1.0

        self.treward += reward

        return self.make_obs(env_obs=o, total_reward=[self.treward]), reward, done, {}


def build_two_goal_maze_env(size_x: int, size_y: int, task_seed: int, n_tasks: int):
    env = MTMiniWorldEnv(
        TwoGoalMazeEnv(
            size_x=size_x, size_y=size_y, task_seed=task_seed, n_tasks=n_tasks
        ),
        task_observation_space=DiscreteSpace(n=1),
    )
    return env



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
"""Wrapper to (lazily) construct a multitask environment from a list of
    constructors (list of functions to construct the environments)."""

from typing import Callable, List, Optional

from gym.core import Env
from gym.spaces.discrete import Discrete as DiscreteSpace

from mtenv import MTEnv
from mtenv.utils import seeding
from mtenv.utils.types import ActionType, EnvObsType, ObsType, StepReturnType

EnvBuilderType = Callable[[], Env]
TaskStateType = int
TaskObsType = int


class MultiEnvWrapper(MTEnv):
    def __init__(
        self,
        funcs_to_make_envs: List[EnvBuilderType],
        initial_task_state: TaskStateType,
    ) -> None:
        """Wrapper to (lazily) construct a multitask environment from a
        list of constructors (list of functions to construct the
        environments).

        The wrapper enables activating/slecting any environment (from the
        list of environments that can be created) and that environment is
        treated as the current task. The environments are created lazily.

        Note that this wrapper is experimental and may change in the future.

        Args:
            funcs_to_make_envs (List[EnvBuilderType]): list of constructor
                functions to make the environments.
            initial_task_state (TaskStateType): intial task/environment
                to select.
        """
        self._num_tasks = len(funcs_to_make_envs)
        self._funcs_to_make_envs = funcs_to_make_envs
        self._envs = [None for _ in range(self._num_tasks)]
        self._envs[initial_task_state] = funcs_to_make_envs[initial_task_state]()
        self.env: Env = self._envs[initial_task_state]
        super().__init__(
            action_space=self.env.action_space,
            env_observation_space=self.env.observation_space,
            task_observation_space=DiscreteSpace(n=self._num_tasks),
        )
        self.task_obs: TaskObsType = initial_task_state

    def _make_observation(self, env_obs: EnvObsType) -> ObsType:
        return {
            "env_obs": env_obs,
            "task_obs": self.task_obs,
        }

    def step(self, action: ActionType) -> StepReturnType:
        env_obs, reward, done, info = self.env.step(action)
        return self._make_observation(env_obs=env_obs), reward, done, info

    def get_task_state(self) -> TaskStateType:
        return self.task_obs

    def set_task_state(self, task_state: TaskStateType) -> None:
        self.task_obs = task_state
        if self._envs[task_state] is None:
            self._envs[task_state] = self._funcs_to_make_envs[task_state]()
        self.env = self._envs[task_state]

    def assert_env_seed_is_set(self) -> None:
        """The seed is set during the call to the constructor of self.env"""
        pass

    def assert_task_seed_is_set(self) -> None:
        assert self.np_random_task is not None, "please call `seed_task()` first"

    def reset(self) -> ObsType:
        return self._make_observation(env_obs=self.env.reset())

    def sample_task_state(self) -> TaskStateType:
        self.assert_task_seed_is_set()
        task_state = self.np_random_task.randint(self._num_tasks)  # type: ignore[union-attr]
        # The assert statement (at the start of the function) ensures that self.np_random_task
        # is not None. Mypy is raising the warning incorrectly.
        assert isinstance(task_state, int)
        return task_state

    def reset_task_state(self) -> None:
        self.set_task_state(task_state=self.sample_task_state())

    def seed(self, seed: Optional[int] = None) -> List[int]:
        self.np_random_env, seed = seeding.np_random(seed)
        env_seeds = self.env.seed(seed)
        if isinstance(env_seeds, list):
            return [seed] + env_seeds
        return [seed]



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for the pixel wrapper."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

# Internal dependencies.
from absl.testing import absltest
from absl.testing import parameterized
from . import cartpole
from dm_control.suite.wrappers import pixels
import dm_env
from dm_env import specs

import numpy as np


class FakePhysics(object):
    def render(self, *args, **kwargs):
        del args
        del kwargs
        return np.zeros((4, 5, 3), dtype=np.uint8)


class FakeArrayObservationEnvironment(dm_env.Environment):
    def __init__(self):
        self.physics = FakePhysics()

    def reset(self):
        return dm_env.restart(np.zeros((2,)))

    def step(self, action):
        del action
        return dm_env.transition(0.0, np.zeros((2,)))

    def action_spec(self):
        pass

    def observation_spec(self):
        return specs.Array(shape=(2,), dtype=np.float)


class PixelsTest(parameterized.TestCase):
    @parameterized.parameters(True, False)
    def test_dict_observation(self, pixels_only):
        pixel_key = "rgb"

        env = cartpole.swingup()

        # Make sure we are testing the right environment for the test.
        observation_spec = env.observation_spec()
        self.assertIsInstance(observation_spec, collections.OrderedDict)

        width = 320
        height = 240

        # The wrapper should only add one observation.
        wrapped = pixels.Wrapper(
            env,
            observation_key=pixel_key,
            pixels_only=pixels_only,
            render_kwargs={"width": width, "height": height},
        )

        wrapped_observation_spec = wrapped.observation_spec()
        self.assertIsInstance(wrapped_observation_spec, collections.OrderedDict)

        if pixels_only:
            self.assertLen(wrapped_observation_spec, 1)
            self.assertEqual([pixel_key], list(wrapped_observation_spec.keys()))
        else:
            expected_length = len(observation_spec) + 1
            self.assertLen(wrapped_observation_spec, expected_length)
            expected_keys = list(observation_spec.keys()) + [pixel_key]
            self.assertEqual(expected_keys, list(wrapped_observation_spec.keys()))

        # Check that the added spec item is consistent with the added observation.
        time_step = wrapped.reset()
        rgb_observation = time_step.observation[pixel_key]
        wrapped_observation_spec[pixel_key].validate(rgb_observation)

        self.assertEqual(rgb_observation.shape, (height, width, 3))
        self.assertEqual(rgb_observation.dtype, np.uint8)

    @parameterized.parameters(True, False)
    def test_single_array_observation(self, pixels_only):
        pixel_key = "depth"

        env = FakeArrayObservationEnvironment()
        observation_spec = env.observation_spec()
        self.assertIsInstance(observation_spec, specs.Array)

        wrapped = pixels.Wrapper(
            env, observation_key=pixel_key, pixels_only=pixels_only
        )
        wrapped_observation_spec = wrapped.observation_spec()
        self.assertIsInstance(wrapped_observation_spec, collections.OrderedDict)

        if pixels_only:
            self.assertLen(wrapped_observation_spec, 1)
            self.assertEqual([pixel_key], list(wrapped_observation_spec.keys()))
        else:
            self.assertLen(wrapped_observation_spec, 2)
            self.assertEqual(
                [pixels.STATE_KEY, pixel_key], list(wrapped_observation_spec.keys())
            )

        time_step = wrapped.reset()

        depth_observation = time_step.observation[pixel_key]
        wrapped_observation_spec[pixel_key].validate(depth_observation)

        self.assertEqual(depth_observation.shape, (4, 5, 3))
        self.assertEqual(depth_observation.dtype, np.uint8)


if __name__ == "__main__":
    absltest.main()



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Reacher domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from . import base
from . import common
from dm_control.suite.utils import randomizers
from dm_control.utils import containers
from dm_control.utils import rewards
import numpy as np

SUITE = containers.TaggedTasks()
_DEFAULT_TIME_LIMIT = 20
_BIG_TARGET = 0.05
_SMALL_TARGET = 0.015


def get_model_and_assets():
    """Returns a tuple containing the model XML string and a dict of assets."""
    return common.read_model("reacher.xml"), common.ASSETS


@SUITE.add("benchmarking", "easy")
def easy(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns reacher with sparse reward with 5e-2 tol and randomized target."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Reacher(target_size=_BIG_TARGET, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics, task, time_limit=time_limit, **environment_kwargs
    )


@SUITE.add("benchmarking")
def hard(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns reacher with sparse reward with 1e-2 tol and randomized target."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Reacher(target_size=_SMALL_TARGET, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics, task, time_limit=time_limit, **environment_kwargs
    )


class Physics(mujoco.Physics):
    """Physics simulation with additional features for the Reacher domain."""

    def finger_to_target(self):
        """Returns the vector from target to finger in global coordinates."""
        return (
            self.named.data.geom_xpos["target", :2]
            - self.named.data.geom_xpos["finger", :2]
        )

    def finger_to_target_dist(self):
        """Returns the signed distance between the finger and target surface."""
        return np.linalg.norm(self.finger_to_target())


class Reacher(base.Task):
    """A reacher `Task` to reach the target."""

    def __init__(self, target_size, random=None):
        """Initialize an instance of `Reacher`.

        Args:
          target_size: A `float`, tolerance to determine whether finger reached the
              target.
          random: Optional, either a `numpy.random.RandomState` instance, an
            integer seed for creating a new `RandomState`, or None to select a seed
            automatically (default).
        """
        self._target_size = target_size
        super(Reacher, self).__init__(random=random)

    def initialize_episode(self, physics):
        """Sets the state of the environment at the start of each episode."""
        physics.named.model.geom_size["target", 0] = self._target_size
        randomizers.randomize_limited_and_rotational_joints(physics, self.random)

        # Randomize target position
        angle = self.random.uniform(0, 2 * np.pi)
        radius = self.random.uniform(0.05, 0.20)
        physics.named.model.geom_pos["target", "x"] = radius * np.sin(angle)
        physics.named.model.geom_pos["target", "y"] = radius * np.cos(angle)

        super(Reacher, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns an observation of the state and the target position."""
        obs = collections.OrderedDict()
        obs["position"] = physics.position()
        obs["to_target"] = physics.finger_to_target()
        obs["velocity"] = physics.velocity()
        return obs

    def get_reward(self, physics):
        radii = physics.named.model.geom_size[["target", "finger"], 0].sum()
        return rewards.tolerance(physics.finger_to_target_dist(), (0, radii))



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
"""Wrapper to fix the number of tasks in an existing multitask environment
and return the id of the task as part of the observation."""

from gym.spaces import Dict as DictSpace
from gym.spaces import Discrete

from mtenv import MTEnv
from mtenv.utils.types import ActionType, ObsType, StepReturnType, TaskStateType
from mtenv.wrappers.ntasks import NTasks


class NTasksId(NTasks):
    def __init__(self, env: MTEnv, n_tasks: int):
        """Wrapper to fix the number of tasks in an existing multitask
        environment to `n_tasks`.

        Each task is sampled in this fixed set of `n_tasks`. The agent
        observes the id of the task.

        Args:
            env (MTEnv): Multitask environment to wrap over.
            n_tasks (int): Number of tasks to sample.
        """
        self.env = env

        super().__init__(n_tasks=n_tasks, env=env)
        self.task_state: TaskStateType
        self.observation_space: DictSpace = DictSpace(
            spaces={
                "env_obs": self.observation_space["env_obs"],
                "task_obs": Discrete(n_tasks),
            }
        )

    def _update_obs(self, obs: ObsType) -> ObsType:
        obs["task_obs"] = self.get_task_obs()
        return obs

    def step(self, action: ActionType) -> StepReturnType:
        obs, reward, done, info = self.env.step(action)
        return self._update_obs(obs), reward, done, info

    def get_task_obs(self) -> TaskStateType:
        return self.task_state

    def get_task_state(self) -> TaskStateType:
        return self.task_state

    def set_task_state(self, task_state: TaskStateType) -> None:
        self.env.set_task_state(self.tasks[task_state])
        self.task_state = task_state

    def reset(self) -> ObsType:
        obs = self.env.reset()
        return self._update_obs(obs)

    def sample_task_state(self) -> TaskStateType:
        self.assert_task_seed_is_set()
        if not self._are_tasks_set:
            self.tasks = [self.env.sample_task_state() for _ in range(self.n_tasks)]
            self._are_tasks_set = True

        # The assert statement (at the start of the function) ensures that self.np_random_task
        # is not None. Mypy is raising the warning incorrectly.
        id_task = self.np_random_task.randint(self.n_tasks)  # type: ignore[union-attr]
        return id_task



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
__version__ = "1.0"

from mtenv.core import MTEnv  # noqa: F401
from mtenv.envs.registration import make  # noqa: F401

__all__ = ["MTEnv", "make"]



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Wrapper control suite environments that adds Gaussian noise to actions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import dm_env
import numpy as np


_BOUNDS_MUST_BE_FINITE = (
    "All bounds in `env.action_spec()` must be finite, got: {action_spec}"
)


class Wrapper(dm_env.Environment):
    """Wraps a control environment and adds Gaussian noise to actions."""

    def __init__(self, env, scale=0.01):
        """Initializes a new action noise Wrapper.

        Args:
          env: The control suite environment to wrap.
          scale: The standard deviation of the noise, expressed as a fraction
            of the max-min range for each action dimension.

        Raises:
          ValueError: If any of the action dimensions of the wrapped environment are
            unbounded.
        """
        action_spec = env.action_spec()
        if not (
            np.all(np.isfinite(action_spec.minimum))
            and np.all(np.isfinite(action_spec.maximum))
        ):
            raise ValueError(_BOUNDS_MUST_BE_FINITE.format(action_spec=action_spec))
        self._minimum = action_spec.minimum
        self._maximum = action_spec.maximum
        self._noise_std = scale * (action_spec.maximum - action_spec.minimum)
        self._env = env

    def step(self, action):
        noisy_action = action + self._env.task.random.normal(scale=self._noise_std)
        # Clip the noisy actions in place so that they fall within the bounds
        # specified by the `action_spec`. Note that MuJoCo implicitly clips out-of-
        # bounds control inputs, but we also clip here in case the actions do not
        # correspond directly to MuJoCo actuators, or if there are other wrapper
        # layers that expect the actions to be within bounds.
        np.clip(noisy_action, self._minimum, self._maximum, out=noisy_action)
        return self._env.step(noisy_action)

    def reset(self):
        return self._env.reset()

    def observation_spec(self):
        return self._env.observation_spec()

    def action_spec(self):
        return self._env.action_spec()

    def __getattr__(self, name):
        return getattr(self._env, name)



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Pendulum domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from . import base
from . import common
from dm_control.utils import containers
from dm_control.utils import rewards
import numpy as np


_DEFAULT_TIME_LIMIT = 20
_ANGLE_BOUND = 8
_COSINE_BOUND = np.cos(np.deg2rad(_ANGLE_BOUND))
SUITE = containers.TaggedTasks()


def get_model_and_assets():
    """Returns a tuple containing the model XML string and a dict of assets."""
    return common.read_model("pendulum.xml"), common.ASSETS


@SUITE.add("benchmarking")
def swingup(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns pendulum swingup task ."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = SwingUp(random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics, task, time_limit=time_limit, **environment_kwargs
    )


class Physics(mujoco.Physics):
    """Physics simulation with additional features for the Pendulum domain."""

    def pole_vertical(self):
        """Returns vertical (z) component of pole frame."""
        return self.named.data.xmat["pole", "zz"]

    def angular_velocity(self):
        """Returns the angular velocity of the pole."""
        return self.named.data.qvel["hinge"].copy()

    def pole_orientation(self):
        """Returns both horizontal and vertical components of pole frame."""
        return self.named.data.xmat["pole", ["zz", "xz"]]


class SwingUp(base.Task):
    """A Pendulum `Task` to swing up and balance the pole."""

    def __init__(self, random=None):
        """Initialize an instance of `Pendulum`.

        Args:
          random: Optional, either a `numpy.random.RandomState` instance, an
            integer seed for creating a new `RandomState`, or None to select a seed
            automatically (default).
        """
        super(SwingUp, self).__init__(random=random)

    def initialize_episode(self, physics):
        """Sets the state of the environment at the start of each episode.

        Pole is set to a random angle between [-pi, pi).

        Args:
          physics: An instance of `Physics`.

        """
        physics.named.data.qpos["hinge"] = self.random.uniform(-np.pi, np.pi)
        super(SwingUp, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns an observation.

        Observations are states concatenating pole orientation and angular velocity
        and pixels from fixed camera.

        Args:
          physics: An instance of `physics`, Pendulum physics.

        Returns:
          A `dict` of observation.
        """
        obs = collections.OrderedDict()
        obs["orientation"] = physics.pole_orientation()
        obs["velocity"] = physics.angular_velocity()
        return obs

    def get_reward(self, physics):
        return rewards.tolerance(physics.pole_vertical(), (_COSINE_BOUND, 1))



# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import scipy.special
from gym import spaces
from gym.utils import seeding

from mtenv import MTEnv


class TMDP(MTEnv):
    """Defines a Tabuular MDP where task_state is the reward matrix,transition matrix
        reward_matrix is n_states*n_actions and gies the probability of having a reward = +1 when choosing action a in state s (matrix[s,a])
        transition_matrix is n_states*n_actions*n_states and gives the probability of moving to state s' when choosing action a in state s (matrix[s,a,s'])
    Args:
        MTEnv ([type]): [description]
    """

    def __init__(self, n_states, n_actions):
        self.n_states = n_states
        self.n_actions = n_actions

        ohigh = np.array([1.0 for n in range(n_states + 1)])
        olow = np.array([0.0 for n in range(n_states + 1)])
        observation_space = spaces.Box(olow, ohigh, dtype=np.float32)
        action_space = spaces.Discrete(n_actions)
        self.task_state = (
            np.zeros((n_states, n_actions)),
            np.zeros((n_states, n_actions, n_states)),
        )
        o = self.get_task_obs()
        thigh = np.ones((len(o),))
        tlow = np.zeros((len(o),))
        task_space = spaces.Box(tlow, thigh, dtype=np.float32)
        super().__init__(
            action_space=action_space,
            env_observation_space=observation_space,
            task_observation_space=task_space,
        )

        # task state is the reward matrix and transition matrix

    def get_task_obs(self):
        obs = list(self.task_state[0].flatten()) + list(self.task_state[1].flatten())
        return obs

    def get_task_state(self):
        return self.task_state

    def set_task_state(self, task_state):
        self.task_state = task_state

    def sample_task_state(self):
        raise NotImplementedError

    def seed(self, env_seed):
        self.np_random_env, seed = seeding.np_random(env_seed)
        return [seed]

    def seed_task(self, task_seed):
        self.np_random_task, seed = seeding.np_random(task_seed)
        return [seed]

    def step(self, action):
        t_reward, t_matrix = self.task_state
        reward = 0.0

        if self.np_random_env.rand() < t_reward[self.state][action]:
            reward = 1.0
        self.state = self.np_random_env.multinomial(
            1, t_matrix[self.state][action]
        ).argmax()

        obs = np.zeros(self.n_states + 1)
        obs[self.state] = 1.0
        obs[-1] = reward
        return (
            {"env_obs": list(obs), "task_obs": self.get_task_obs()},
            reward,
            False,
            {},
        )

    def reset(self):
        self.state = self.np_random_env.randint(self.n_states)
        obs = np.zeros(self.n_states + 1)
        obs[self.state] = 1.0
        return {"env_obs": list(obs), "task_obs": self.get_task_obs()}


class UniformTMDP(TMDP):
    def __init__(self, n_states, n_actions):
        super().__init__(n_states, n_actions)

    def sample_task_state(self):
        self.assert_task_seed_is_set()
        t_reward = self.np_random_task.rand(self.n_states, self.n_actions)
        t_transitions = self.np_random_task.randn(
            self.n_states, self.n_actions, self.n_states
        )
        t_transitions = scipy.special.softmax(t_transitions, axis=2)

        new_task_state = t_reward, t_transitions
        return new_task_state


if __name__ == "__main__":
    env = UniformTMDP(3, 2)
    env.seed(5)
    env.seed_task(14)
    env.reset_task_state()
    obs = env.reset()
    done = False
    while not done:
        action = np.random.randint(env.action_space.n)
        obs, rew, done, _ = env.step(action)
        print(obs["env_obs"])



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

r"""Optimal policy for LQR levels.

LQR control problem is described in
https://en.wikipedia.org/wiki/Linear-quadratic_regulator#Infinite-horizon.2C_discrete-time_LQR
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import logging
from dm_control.mujoco import wrapper
import numpy as np
from six.moves import range

try:
    import scipy.linalg as sp  # pylint: disable=g-import-not-at-top
except ImportError:
    sp = None


def _solve_dare(a, b, q, r):
    """Solves the Discrete-time Algebraic Riccati Equation (DARE) by iteration.

    Algebraic Riccati Equation:
    ```none
    P_{t-1} = Q + A' * P_{t} * A -
              A' * P_{t} * B * (R + B' * P_{t} * B)^{-1} * B' * P_{t} * A
    ```

    Args:
      a: A 2 dimensional numpy array, transition matrix A.
      b: A 2 dimensional numpy array, control matrix B.
      q: A 2 dimensional numpy array, symmetric positive definite cost matrix.
      r: A 2 dimensional numpy array, symmetric positive definite cost matrix

    Returns:
      A numpy array, a real symmetric matrix P which is the solution to DARE.

    Raises:
      RuntimeError: If the computed P matrix is not symmetric and
        positive-definite.
    """
    p = np.eye(len(a))
    for _ in range(1000000):
        a_p = a.T.dot(p)  # A' * P_t
        a_p_b = np.dot(a_p, b)  # A' * P_t * B
        # Algebraic Riccati Equation.
        p_next = (
            q
            + np.dot(a_p, a)
            - a_p_b.dot(np.linalg.solve(b.T.dot(p.dot(b)) + r, a_p_b.T))
        )
        p_next += p_next.T
        p_next *= 0.5
        if np.abs(p - p_next).max() < 1e-12:
            break
        p = p_next
    else:
        logging.warning("DARE solver did not converge")
    try:
        # Check that the result is symmetric and positive-definite.
        np.linalg.cholesky(p_next)
    except np.linalg.LinAlgError:
        raise RuntimeError(
            "ARE solver failed: P matrix is not symmetric and " "positive-definite."
        )
    return p_next


def solve(env):
    """Returns the optimal value and policy for LQR problem.

    Args:
      env: An instance of `control.EnvironmentV2` with LQR level.

    Returns:
      p: A numpy array, the Hessian of the optimal total cost-to-go (value
        function at state x) is V(x) = .5 * x' * p * x.
      k: A numpy array which gives the optimal linear policy u = k * x.
      beta: The maximum eigenvalue of (a + b * k). Under optimal policy, at
        timestep n the state tends to 0 like beta^n.

    Raises:
      RuntimeError: If the controlled system is unstable.
    """
    n = env.physics.model.nq  # number of DoFs
    m = env.physics.model.nu  # number of controls

    # Compute the mass matrix.
    mass = np.zeros((n, n))
    wrapper.mjbindings.mjlib.mj_fullM(env.physics.model.ptr, mass, env.physics.data.qM)

    # Compute input matrices a, b, q and r to the DARE solvers.
    # State transition matrix a.
    stiffness = np.diag(env.physics.model.jnt_stiffness.ravel())
    damping = np.diag(env.physics.model.dof_damping.ravel())
    dt = env.physics.model.opt.timestep

    j = np.linalg.solve(-mass, np.hstack((stiffness, damping)))
    a = np.eye(2 * n) + dt * np.vstack(
        (dt * j + np.hstack((np.zeros((n, n)), np.eye(n))), j)
    )

    # Control transition matrix b.
    b = env.physics.data.actuator_moment.T
    bc = np.linalg.solve(mass, b)
    b = dt * np.vstack((dt * bc, bc))

    # State cost Hessian q.
    q = np.diag(np.hstack([np.ones(n), np.zeros(n)]))

    # Control cost Hessian r.
    r = env.task.control_cost_coef * np.eye(m)

    if sp:
        # Use scipy's faster DARE solver if available.
        solve_dare = sp.solve_discrete_are
    else:
        # Otherwise fall back on a slower internal implementation.
        solve_dare = _solve_dare

    # Solve the discrete algebraic Riccati equation.
    p = solve_dare(a, b, q, r)
    k = -np.linalg.solve(b.T.dot(p.dot(b)) + r, b.T.dot(p.dot(a)))

    # Under optimal policy, state tends to 0 like beta^n_timesteps
    beta = np.abs(np.linalg.eigvals(a + b.dot(k))).max()
    if beta >= 1.0:
        raise RuntimeError("Controlled system is unstable.")
    return p, k, beta



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import math

import numpy as np
from gym import logger, spaces

from mtenv import MTEnv
from mtenv.utils import seeding

"""
Classic cart-pole system implemented based on Rich Sutton et al.
Copied from http://incompleteideas.net/sutton/book/code/pole.c
permalink: https://perma.cc/C9ZM-652R
"""


class MTCartPole(MTEnv):
    """A cartpole environment with varying physical values
    (see the self._mu_to_vars function)
    """

    metadata = {"render.modes": ["human", "rgb_array"], "video.frames_per_second": 50}

    def _mu_to_vars(self, mu):
        self.gravity = 9.8 + mu[0] * 5
        self.masscart = 1.0 + mu[1] * 0.5
        self.masspole = 0.1 + mu[2] * 0.09
        self.total_mass = self.masspole + self.masscart
        self.length = 0.5 + mu[3] * 0.3
        self.polemass_length = self.masspole * self.length
        self.force_mag = 10 * mu[4]
        if mu[4] == 0:
            self.force_mag = 10

    def __init__(self):
        # Angle limit set to 2 * theta_threshold_radians so failing observation is still within bounds
        self.x_threshold = 2.4
        self.theta_threshold_radians = 12 * 2 * math.pi / 360

        high = np.array(
            [
                self.x_threshold * 2,
                np.finfo(np.float32).max,
                self.theta_threshold_radians * 2,
                np.finfo(np.float32).max,
            ]
        )
        observation_space = spaces.Box(-high, high, dtype=np.float32)
        action_space = spaces.Discrete(2)
        high = np.array([1.0 for k in range(5)])
        task_space = spaces.Box(-high, high, dtype=np.float32)
        super().__init__(
            action_space=action_space,
            env_observation_space=observation_space,
            task_observation_space=task_space,
        )

        self.gravity = 9.8
        self.masscart = 1.0
        self.masspole = 0.1
        self.total_mass = self.masspole + self.masscart
        self.length = 0.5  # actually half the pole's length
        self.polemass_length = self.masspole * self.length
        self.force_mag = 10.0
        self.tau = 0.02  # seconds between state updates
        self.kinematics_integrator = "euler"
        # Angle at which to fail the episode

        self.state = None
        self.steps_beyond_done = None

        self.task_state = None

    def step(self, action):
        self.t += 1
        self._mu_to_vars(self.task_state)

        assert self.action_space.contains(action), "%r (%s) invalid" % (
            action,
            type(action),
        )
        state = self.state
        x, x_dot, theta, theta_dot = state
        force = self.force_mag if action == 1 else -self.force_mag
        costheta = math.cos(theta)
        sintheta = math.sin(theta)
        temp = (
            force + self.polemass_length * theta_dot * theta_dot * sintheta
        ) / self.total_mass
        thetaacc = (self.gravity * sintheta - costheta * temp) / (
            self.length
            * (4.0 / 3.0 - self.masspole * costheta * costheta / self.total_mass)
        )
        xacc = temp - self.polemass_length * thetaacc * costheta / self.total_mass
        if self.kinematics_integrator == "euler":
            x = x + self.tau * x_dot
            x_dot = x_dot + self.tau * xacc
            theta = theta + self.tau * theta_dot
            theta_dot = theta_dot + self.tau * thetaacc
        else:  # semi-implicit euler
            x_dot = x_dot + self.tau * xacc
            x = x + self.tau * x_dot
            theta_dot = theta_dot + self.tau * thetaacc
            theta = theta + self.tau * theta_dot
        self.state = [x, x_dot, theta, theta_dot]
        done = (
            x < -self.x_threshold
            or x > self.x_threshold
            or theta < -self.theta_threshold_radians
            or theta > self.theta_threshold_radians
        )
        done = bool(done)

        reward = 0

        if not done:
            reward = 1.0
        elif self.steps_beyond_done is None:
            # Pole just fell!
            self.steps_beyond_done = 0
            reward = 1.0
        else:
            if self.steps_beyond_done == 0:
                logger.warn(
                    "You are calling 'step()' even though this environment has already returned done = True. You should always call 'reset()' once you receive 'done = True' -- any further steps are undefined behavior."
                )
                print(
                    "You are calling 'step()' even though this environment has already returned done = True. You should always call 'reset()' once you receive 'done = True' -- any further steps are undefined behavior."
                )
            self.steps_beyond_done += 1
            reward = 0.0

        return (
            {"env_obs": self.state, "task_obs": self.get_task_obs()},
            reward,
            done,
            {},
        )

    def reset(self, **args):
        self.assert_env_seed_is_set()
        assert self.task_state is not None

        self._mu_to_vars(self.task_state)
        self.state = self.np_random_env.uniform(low=-0.05, high=0.05, size=(4,))
        self.steps_beyond_done = None
        self.t = 0
        return {"env_obs": self.state, "task_obs": self.get_task_obs()}

    def get_task_obs(self):
        return self.task_state

    def get_task_state(self):
        return self.task_state

    def set_task_state(self, task_state):
        self.task_state = task_state

    def sample_task_state(self):
        self.assert_task_seed_is_set()
        super().sample_task_state()
        new_task_state = [
            self.np_random_task.uniform(-1, 1),
            self.np_random_task.uniform(-1, 1),
            self.np_random_task.uniform(-1, 1),
            self.np_random_task.uniform(-1, 1),
            self.np_random_task.uniform(-1, 1),
        ]
        return new_task_state

    def seed(self, env_seed):
        self.np_random_env, seed = seeding.np_random(env_seed)
        return [seed]

    def seed_task(self, task_seed):
        self.np_random_task, seed = seeding.np_random(task_seed)
        return [seed]


class CartPole(MTCartPole):
    """The original cartpole environment in the MTEnv fashion"""

    def __init__(self):
        super().__init__()

    def sample_task_state(self):
        new_task_state = [0.0, 0.0, 0.0, 0.0, 0.0]
        return new_task_state


if __name__ == "__main__":
    env = MTCartPole()
    env.seed(5)
    env.seed_task(15)
    env.reset_task_state()
    obs = env.reset()
    print(obs)
    done = False
    while not done:
        obs, rew, done, _ = env.step(np.random.randint(env.action_space.n))
        print(obs)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
"""Core API of MultiTask Environments for Reinforcement Learning."""
from abc import ABC, abstractmethod
from typing import List, Optional

from gym.core import Env
from gym.spaces.dict import Dict as DictSpace
from gym.spaces.space import Space
from numpy.random import RandomState

from mtenv.utils import seeding
from mtenv.utils.types import (
    ActionType,
    ObsType,
    StepReturnType,
    TaskObsType,
    TaskStateType,
)


class MTEnv(Env, ABC):  # type: ignore[misc]
    def __init__(
        self,
        action_space: Space,
        env_observation_space: Space,
        task_observation_space: Space,
    ) -> None:
        """Main class for multitask RL Environments.

        This abstract class extends the OpenAI Gym environment and adds
        support for return the task-specific information from the environment.
        The observation returned from the single task environments is
        encoded as `env_obs` (environment observation) while the task
        specific observation is encoded as the `task_obs` (task observation).
        The observation returned by `mtenv` is a dictionary of `env_obs` and
        `task_obs`.  Since this class extends the OpenAI gym, the `mtenv`
        API looks similar to the gym API.

        .. code-block:: python

            import mtenv
            env = mtenv.make('xxx')
            env.reset()

        Any multitask RL environment class should extend/implement this class.

        Args:
            action_space (Space)
            env_observation_space (Space)
            task_observation_space (Space)
        """

        self.action_space = action_space
        self.observation_space: DictSpace = DictSpace(
            spaces={
                "env_obs": env_observation_space,
                "task_obs": task_observation_space,
            }
        )

        self.np_random_env: Optional[RandomState] = None
        self.np_random_task: Optional[RandomState] = None

        self._task_obs: TaskObsType

    @abstractmethod
    def step(self, action: ActionType) -> StepReturnType:
        """Execute the action in the environment.

        Args:
            action (ActionType)

        Returns:
            StepReturnType: Tuple of `multitask observation`, `reward`,
            `done`, and `info`. For more information on `multitask observation`
            returned by the environment, refer :ref:`multitask_observation`.
        """
        pass

    def get_task_obs(self) -> TaskObsType:
        """Get the current value of task observation.

        Environment returns task observation everytime we call `step` or
        `reset`. This function is useful when the user wants to access the
        task observation without acting in (or resetting) the environment.

        Returns:
            TaskObsType:
        """
        return self._task_obs

    @abstractmethod
    def get_task_state(self) -> TaskStateType:
        """Return all the information needed to execute the current task
        again.

        This function is useful when we want to set the environment to a
        previous task.

        Returns:
            TaskStateType: For more information on `task_state`, refer :ref:`task_state`.
        """
        pass

    @abstractmethod
    def set_task_state(self, task_state: TaskStateType) -> None:
        """Reset the environment to a particular task.

        `task_state` contains all the information that the environment
        needs to switch to any other task.

        Args:
            task_state (TaskStateType): For more information on `task_state`,
                refer :ref:`task_state`.
        """
        pass

    def assert_env_seed_is_set(self) -> None:
        """Check that seed (for the environment) is set.

        `reset` function should invoke this function before resetting the
        environment (for reproducibility).

        """
        assert self.np_random_env is not None, "please call `seed()` first"

    def assert_task_seed_is_set(self) -> None:
        """Check that seed (for the task) is set.

        `sample_task_state` function should invoke this function before
        sampling a new task state (for reproducibility).

        """
        assert self.np_random_task is not None, "please call `seed_task()` first"

    @abstractmethod
    def reset(self) -> ObsType:
        """Reset the environment to some initial state and return the
        observation in the new state.

        The subclasses, extending this class, should ensure that the
        environment seed is set (by calling `seed(int)`) before invoking this
        method (for reproducibility). It can be done by invoking
        `self.assert_env_seed_is_set()`.

        Returns:
            ObsType: For more information on `multitask observation`
            returned by the environment, refer :ref:`multitask_observation`.
        """
        pass

    @abstractmethod
    def sample_task_state(self) -> TaskStateType:
        """Sample a `task_state`.

        `task_state` contains all the information that the environment
        needs to switch to any other task.

        The subclasses, extending this class, should ensure that the task
        seed is set (by calling `seed(int)`) before invoking this
        method (for reproducibility). It can be done by invoking
        `self.assert_task_seed_is_set()`.

        Returns:
            TaskStateType: For more information on `task_state`,
            refer :ref:`task_state`.
        """
        pass

    def reset_task_state(self) -> None:
        """Sample a new task_state and set the environment to that `task_state`.

        For more information on `task_state`, refer :ref:`task_state`.
        """
        self.set_task_state(task_state=self.sample_task_state())

    def seed(self, seed: Optional[int] = None) -> List[int]:
        """Set the seed for the environment's random number generator.

        Invoke `seed_task` to set the seed for the task's
        random number generator.

        Args:
            seed (Optional[int], optional): Defaults to None.

        Returns:
            List[int]: Returns the list of seeds used in the environment's
            random number generator. The first value in the list should be
            the seed that should be passed to this method for reproducibility.
        """
        self.np_random_env, seed = seeding.np_random(seed)
        assert isinstance(seed, int)
        return [seed]

    def seed_task(self, seed: Optional[int] = None) -> List[int]:
        """Set the seed for the task's random number generator.

        Invoke `seed` to set the seed for the environment's
        random number generator.

        Args:
            seed (Optional[int], optional): Defaults to None.

        Returns:
            List[int]: Returns the list of seeds used in the task's
            random number generator. The first value in the list should be
            the seed that should be passed to this method for reproducibility.
        """
        self.np_random_task, seed = seeding.np_random(seed)
        assert isinstance(seed, int)
        self.observation_space["task_obs"].seed(seed)
        return [seed]



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for the action noise wrapper."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Internal dependencies.
from absl.testing import absltest
from absl.testing import parameterized
from dm_control.rl import control
from dm_control.suite.wrappers import action_noise
from dm_env import specs
import mock
import numpy as np


class ActionNoiseTest(parameterized.TestCase):
    def make_action_spec(self, lower=(-1.0,), upper=(1.0,)):
        lower, upper = np.broadcast_arrays(lower, upper)
        return specs.BoundedArray(
            shape=lower.shape, dtype=float, minimum=lower, maximum=upper
        )

    def make_mock_env(self, action_spec=None):
        action_spec = action_spec or self.make_action_spec()
        env = mock.Mock(spec=control.Environment)
        env.action_spec.return_value = action_spec
        return env

    def assertStepCalledOnceWithCorrectAction(self, env, expected_action):
        # NB: `assert_called_once_with()` doesn't support numpy arrays.
        env.step.assert_called_once()
        actual_action = env.step.call_args_list[0][0][0]
        np.testing.assert_array_equal(expected_action, actual_action)

    @parameterized.parameters(
        [
            dict(lower=np.r_[-1.0, 0.0], upper=np.r_[1.0, 2.0], scale=0.05),
            dict(lower=np.r_[-1.0, 0.0], upper=np.r_[1.0, 2.0], scale=0.0),
            dict(lower=np.r_[-1.0, 0.0], upper=np.r_[-1.0, 0.0], scale=0.05),
        ]
    )
    def test_step(self, lower, upper, scale):
        seed = 0
        std = scale * (upper - lower)
        expected_noise = np.random.RandomState(seed).normal(scale=std)
        action = np.random.RandomState(seed).uniform(lower, upper)
        expected_noisy_action = np.clip(action + expected_noise, lower, upper)
        task = mock.Mock(spec=control.Task)
        task.random = np.random.RandomState(seed)
        action_spec = self.make_action_spec(lower=lower, upper=upper)
        env = self.make_mock_env(action_spec=action_spec)
        env.task = task
        wrapped_env = action_noise.Wrapper(env, scale=scale)
        time_step = wrapped_env.step(action)
        self.assertStepCalledOnceWithCorrectAction(env, expected_noisy_action)
        self.assertIs(time_step, env.step(expected_noisy_action))

    @parameterized.named_parameters(
        [
            dict(testcase_name="within_bounds", action=np.r_[-1.0], noise=np.r_[0.1]),
            dict(testcase_name="below_lower", action=np.r_[-1.0], noise=np.r_[-0.1]),
            dict(testcase_name="above_upper", action=np.r_[1.0], noise=np.r_[0.1]),
        ]
    )
    def test_action_clipping(self, action, noise):
        lower = -1.0
        upper = 1.0
        expected_noisy_action = np.clip(action + noise, lower, upper)
        task = mock.Mock(spec=control.Task)
        task.random = mock.Mock(spec=np.random.RandomState)
        task.random.normal.return_value = noise
        action_spec = self.make_action_spec(lower=lower, upper=upper)
        env = self.make_mock_env(action_spec=action_spec)
        env.task = task
        wrapped_env = action_noise.Wrapper(env)
        time_step = wrapped_env.step(action)
        self.assertStepCalledOnceWithCorrectAction(env, expected_noisy_action)
        self.assertIs(time_step, env.step(expected_noisy_action))

    @parameterized.parameters(
        [
            dict(lower=np.r_[-1.0, 0.0], upper=np.r_[1.0, np.inf]),
            dict(lower=np.r_[np.nan, 0.0], upper=np.r_[1.0, 2.0]),
        ]
    )
    def test_error_if_action_bounds_non_finite(self, lower, upper):
        action_spec = self.make_action_spec(lower=lower, upper=upper)
        env = self.make_mock_env(action_spec=action_spec)
        with self.assertRaisesWithLiteralMatch(
            ValueError,
            action_noise._BOUNDS_MUST_BE_FINITE.format(action_spec=action_spec),
        ):
            _ = action_noise.Wrapper(env)

    def test_reset(self):
        env = self.make_mock_env()
        wrapped_env = action_noise.Wrapper(env)
        time_step = wrapped_env.reset()
        env.reset.assert_called_once_with()
        self.assertIs(time_step, env.reset())

    def test_observation_spec(self):
        env = self.make_mock_env()
        wrapped_env = action_noise.Wrapper(env)
        observation_spec = wrapped_env.observation_spec()
        env.observation_spec.assert_called_once_with()
        self.assertIs(observation_spec, env.observation_spec())

    def test_action_spec(self):
        env = self.make_mock_env()
        wrapped_env = action_noise.Wrapper(env)
        # `env.action_spec()` is called in `Wrapper.__init__()`
        env.action_spec.reset_mock()
        action_spec = wrapped_env.action_spec()
        env.action_spec.assert_called_once_with()
        self.assertIs(action_spec, env.action_spec())

    @parameterized.parameters(["task", "physics", "control_timestep"])
    def test_getattr(self, attribute_name):
        env = self.make_mock_env()
        wrapped_env = action_noise.Wrapper(env)
        attr = getattr(wrapped_env, attribute_name)
        self.assertIs(attr, getattr(env, attribute_name))


if __name__ == "__main__":
    absltest.main()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
"""Wrapper to convert an environment into multitask environment."""
from typing import Any, Dict, List, Optional

from gym.core import Env
from gym.spaces.space import Space

from mtenv import MTEnv
from mtenv.utils import seeding
from mtenv.utils.types import (
    ActionType,
    EnvObsType,
    ObsType,
    StepReturnType,
    TaskObsType,
    TaskStateType,
)


class EnvToMTEnv(MTEnv):
    def __init__(self, env: Env, task_observation_space: Space) -> None:
        """Wrapper to convert an environment into a multitak environment.

        Args:
            env (Env): Environment to wrap over.
            task_observation_space (Space): Task observation space for the
                resulting multitask environment.
        """

        super().__init__(
            action_space=env.action_space,
            env_observation_space=env.observation_space,
            task_observation_space=task_observation_space,
        )

        self.env = env
        self.reward_range = self.env.reward_range
        self.metadata = self.env.metadata

    @property
    def spec(self) -> Any:
        return self.env.spec

    @classmethod
    def class_name(cls) -> str:
        return cls.__name__

    def _make_observation(self, env_obs: EnvObsType) -> ObsType:
        return {"env_obs": env_obs, "task_obs": self.get_task_obs()}

    def get_task_obs(self) -> TaskObsType:
        return self._task_obs

    def get_task_state(self) -> TaskStateType:
        raise NotImplementedError

    def set_task_state(self, task_state: TaskStateType) -> None:
        raise NotImplementedError

    def sample_task_state(self) -> TaskStateType:
        raise NotImplementedError

    def reset(self, **kwargs: Dict[str, Any]) -> ObsType:
        self.assert_env_seed_is_set()
        env_obs = self.env.reset(**kwargs)
        return self._make_observation(env_obs=env_obs)

    def reset_task_state(self) -> None:
        self.set_task_state(task_state=self.sample_task_state())

    def step(self, action: ActionType) -> StepReturnType:
        env_obs, reward, done, info = self.env.step(action)
        return (
            self._make_observation(env_obs=env_obs),
            reward,
            done,
            info,
        )

    def seed(self, seed: Optional[int] = None) -> List[int]:
        self.np_random_env, seed = seeding.np_random(seed)
        env_seeds = self.env.seed(seed)
        if isinstance(env_seeds, list):
            return [seed] + env_seeds
        return [seed]

    def render(self, mode: str = "human", **kwargs: Dict[str, Any]) -> Any:
        """Renders the environment."""
        return self.env.render(mode, **kwargs)

    def close(self) -> Any:
        return self.env.close()

    def __str__(self) -> str:
        return f"{type(self).__name__}{self.env}"

    def __repr__(self) -> str:
        return str(self)

    @property
    def unwrapped(self) -> Env:
        return self.env.unwrapped

    def __getattr__(self, name: str) -> Any:
        if name.startswith("_"):
            raise AttributeError(
                "attempted to get missing private attribute '{}'".format(name)
            )
        return getattr(self.env, name)



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Procedurally generated LQR domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import os

from dm_control import mujoco
from dm_control.rl import control
from . import base
from . import common
from dm_control.utils import containers
from dm_control.utils import xml_tools
from lxml import etree
import numpy as np
from six.moves import range

from dm_control.utils import io as resources

_DEFAULT_TIME_LIMIT = float("inf")
_CONTROL_COST_COEF = 0.1
SUITE = containers.TaggedTasks()


def get_model_and_assets(n_bodies, n_actuators, random):
    """Returns the model description as an XML string and a dict of assets.

    Args:
      n_bodies: An int, number of bodies of the LQR.
      n_actuators: An int, number of actuated bodies of the LQR. `n_actuators`
        should be less or equal than `n_bodies`.
      random: A `numpy.random.RandomState` instance.

    Returns:
      A tuple `(model_xml_string, assets)`, where `assets` is a dict consisting of
      `{filename: contents_string}` pairs.
    """
    return _make_model(n_bodies, n_actuators, random), common.ASSETS


@SUITE.add()
def lqr_2_1(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns an LQR environment with 2 bodies of which the first is actuated."""
    return _make_lqr(
        n_bodies=2,
        n_actuators=1,
        control_cost_coef=_CONTROL_COST_COEF,
        time_limit=time_limit,
        random=random,
        environment_kwargs=environment_kwargs,
    )


@SUITE.add()
def lqr_6_2(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns an LQR environment with 6 bodies of which first 2 are actuated."""
    return _make_lqr(
        n_bodies=6,
        n_actuators=2,
        control_cost_coef=_CONTROL_COST_COEF,
        time_limit=time_limit,
        random=random,
        environment_kwargs=environment_kwargs,
    )


def _make_lqr(
    n_bodies, n_actuators, control_cost_coef, time_limit, random, environment_kwargs
):
    """Returns a LQR environment.

    Args:
      n_bodies: An int, number of bodies of the LQR.
      n_actuators: An int, number of actuated bodies of the LQR. `n_actuators`
        should be less or equal than `n_bodies`.
      control_cost_coef: A number, the coefficient of the control cost.
      time_limit: An int, maximum time for each episode in seconds.
      random: Either an existing `numpy.random.RandomState` instance, an
        integer seed for creating a new `RandomState`, or None to select a seed
        automatically.
      environment_kwargs: A `dict` specifying keyword arguments for the
        environment, or None.

    Returns:
      A LQR environment with `n_bodies` bodies of which first `n_actuators` are
      actuated.
    """

    if not isinstance(random, np.random.RandomState):
        random = np.random.RandomState(random)

    model_string, assets = get_model_and_assets(n_bodies, n_actuators, random=random)
    physics = Physics.from_xml_string(model_string, assets=assets)
    task = LQRLevel(control_cost_coef, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics, task, time_limit=time_limit, **environment_kwargs
    )


def _make_body(body_id, stiffness_range, damping_range, random):
    """Returns an `etree.Element` defining a body.

    Args:
      body_id: Id of the created body.
      stiffness_range: A tuple of (stiffness_lower_bound, stiffness_uppder_bound).
        The stiffness of the joint is drawn uniformly from this range.
      damping_range: A tuple of (damping_lower_bound, damping_upper_bound). The
        damping of the joint is drawn uniformly from this range.
      random: A `numpy.random.RandomState` instance.

    Returns:
     A new instance of `etree.Element`. A body element with two children: joint
     and geom.
    """
    body_name = "body_{}".format(body_id)
    joint_name = "joint_{}".format(body_id)
    geom_name = "geom_{}".format(body_id)

    body = etree.Element("body", name=body_name)
    body.set("pos", ".25 0 0")
    joint = etree.SubElement(body, "joint", name=joint_name)
    body.append(etree.Element("geom", name=geom_name))
    joint.set("stiffness", str(random.uniform(stiffness_range[0], stiffness_range[1])))
    joint.set("damping", str(random.uniform(damping_range[0], damping_range[1])))
    return body


def _make_model(
    n_bodies, n_actuators, random, stiffness_range=(15, 25), damping_range=(0, 0)
):
    """Returns an MJCF XML string defining a model of springs and dampers.

    Args:
      n_bodies: An integer, the number of bodies (DoFs) in the system.
      n_actuators: An integer, the number of actuated bodies.
      random: A `numpy.random.RandomState` instance.
      stiffness_range: A tuple containing minimum and maximum stiffness. Each
        joint's stiffness is sampled uniformly from this interval.
      damping_range: A tuple containing minimum and maximum damping. Each joint's
        damping is sampled uniformly from this interval.

    Returns:
      An MJCF string describing the linear system.

    Raises:
      ValueError: If the number of bodies or actuators is erronous.
    """
    if n_bodies < 1 or n_actuators < 1:
        raise ValueError("At least 1 body and 1 actuator required.")
    if n_actuators > n_bodies:
        raise ValueError("At most 1 actuator per body.")

    file_path = os.path.join(os.path.dirname(__file__), "lqr.xml")
    with resources.GetResourceAsFile(file_path) as xml_file:
        mjcf = xml_tools.parse(xml_file)
    parent = mjcf.find("./worldbody")
    actuator = etree.SubElement(mjcf.getroot(), "actuator")
    tendon = etree.SubElement(mjcf.getroot(), "tendon")

    for body in range(n_bodies):
        # Inserting body.
        child = _make_body(body, stiffness_range, damping_range, random)
        site_name = "site_{}".format(body)
        child.append(etree.Element("site", name=site_name))

        if body == 0:
            child.set("pos", ".25 0 .1")
        # Add actuators to the first n_actuators bodies.
        if body < n_actuators:
            # Adding actuator.
            joint_name = "joint_{}".format(body)
            motor_name = "motor_{}".format(body)
            child.find("joint").set("name", joint_name)
            actuator.append(etree.Element("motor", name=motor_name, joint=joint_name))

        # Add a tendon between consecutive bodies (for visualisation purposes only).
        if body < n_bodies - 1:
            child_site_name = "site_{}".format(body + 1)
            tendon_name = "tendon_{}".format(body)
            spatial = etree.SubElement(tendon, "spatial", name=tendon_name)
            spatial.append(etree.Element("site", site=site_name))
            spatial.append(etree.Element("site", site=child_site_name))
        parent.append(child)
        parent = child

    return etree.tostring(mjcf, pretty_print=True)


class Physics(mujoco.Physics):
    """Physics simulation with additional features for the LQR domain."""

    def state_norm(self):
        """Returns the norm of the physics state."""
        return np.linalg.norm(self.state())


class LQRLevel(base.Task):
    """A Linear Quadratic Regulator `Task`."""

    _TERMINAL_TOL = 1e-6

    def __init__(self, control_cost_coef, random=None):
        """Initializes an LQR level with cost = sum(states^2) + c*sum(controls^2).

        Args:
          control_cost_coef: The coefficient of the control cost.
          random: Optional, either a `numpy.random.RandomState` instance, an
            integer seed for creating a new `RandomState`, or None to select a seed
            automatically (default).

        Raises:
          ValueError: If the control cost coefficient is not positive.
        """
        if control_cost_coef <= 0:
            raise ValueError("control_cost_coef must be positive.")

        self._control_cost_coef = control_cost_coef
        super(LQRLevel, self).__init__(random=random)

    @property
    def control_cost_coef(self):
        return self._control_cost_coef

    def initialize_episode(self, physics):
        """Random state sampled from a unit sphere."""
        ndof = physics.model.nq
        unit = self.random.randn(ndof)
        physics.data.qpos[:] = np.sqrt(2) * unit / np.linalg.norm(unit)
        super(LQRLevel, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns an observation of the state."""
        obs = collections.OrderedDict()
        obs["position"] = physics.position()
        obs["velocity"] = physics.velocity()
        return obs

    def get_reward(self, physics):
        """Returns a quadratic state and control reward."""
        position = physics.position()
        state_cost = 0.5 * np.dot(position, position)
        control_signal = physics.control()
        control_l2_norm = 0.5 * np.dot(control_signal, control_signal)
        return 1 - (state_cost + control_l2_norm * self._control_cost_coef)

    def get_evaluation(self, physics):
        """Returns a sparse evaluation reward that is not used for learning."""
        return float(physics.state_norm() <= 0.01)

    def get_termination(self, physics):
        """Terminates when the state norm is smaller than epsilon."""
        if physics.state_norm() < self._TERMINAL_TOL:
            return 0.0



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Control suite environments explorer."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app
from absl import flags
from dm_control import suite
from dm_control.suite.wrappers import action_noise
from six.moves import input

from dm_control import viewer


_ALL_NAMES = [".".join(domain_task) for domain_task in suite.ALL_TASKS]

flags.DEFINE_enum(
    "environment_name",
    None,
    _ALL_NAMES,
    "Optional 'domain_name.task_name' pair specifying the "
    "environment to load. If unspecified a prompt will appear to "
    "select one.",
)
flags.DEFINE_bool("timeout", True, "Whether episodes should have a time limit.")
flags.DEFINE_bool(
    "visualize_reward",
    True,
    "Whether to vary the colors of geoms according to the " "current reward value.",
)
flags.DEFINE_float(
    "action_noise",
    0.0,
    "Standard deviation of Gaussian noise to apply to actions, "
    "expressed as a fraction of the max-min range for each "
    "action dimension. Defaults to 0, i.e. no noise.",
)
FLAGS = flags.FLAGS


def prompt_environment_name(prompt, values):
    environment_name = None
    while not environment_name:
        environment_name = input(prompt)
        if not environment_name or values.index(environment_name) < 0:
            print('"%s" is not a valid environment name.' % environment_name)
            environment_name = None
    return environment_name


def main(argv):
    del argv
    environment_name = FLAGS.environment_name
    if environment_name is None:
        print("\n  ".join(["Available environments:"] + _ALL_NAMES))
        environment_name = prompt_environment_name(
            "Please select an environment name: ", _ALL_NAMES
        )

    index = _ALL_NAMES.index(environment_name)
    domain_name, task_name = suite.ALL_TASKS[index]

    task_kwargs = {}
    if not FLAGS.timeout:
        task_kwargs["time_limit"] = float("inf")

    def loader():
        env = suite.load(
            domain_name=domain_name, task_name=task_name, task_kwargs=task_kwargs
        )
        env.task.visualize_reward = FLAGS.visualize_reward
        if FLAGS.action_noise > 0:
            env = action_noise.Wrapper(env, scale=FLAGS.action_noise)
        return env

    viewer.launch(loader)


if __name__ == "__main__":
    app.run(main)
